#!/usr/bin/perl -w

use strict;
use Cwd;
use File::Basename;
use File::Copy;
use XML::Simple;
use Getopt::Mixed "nextOption";
use Data::Dumper;

my $ds      = '/';
my $os_name = '';

if($^O eq 'MSWin32') {
	$ds = '\\';
	
	require Win32;
	import  Win32;

	$os_name = Win32::GetOSName();
}

my $dir             = dirname($0);
   $dir            .= $ds;

# Installation on Linux is assumed
my $prefix          = '/usr/local/';
my $wwwroot         = '/var/www/';
my $dir_mode        = 0775;

my $bin_dir         = $prefix.'bin/'; # Linux assumed
my $conf_dir        = $prefix.'etc/fci/';
my $data_dir        = $prefix.'share/fci/';
my $www_dir         = $wwwroot.'fci/';

my $owner           = 'root';
my $group           = 'www-data';

my $qd              = 'qd';

my $apache_conf_src = $dir.'apache-conf';
my $apache_conf_dst = '/etc/apache2/sites-available/fci';	# Default on Debian/Ubuntu

my $lsb_release     = '/bin/lsb_release';

if($^O eq 'linux') {
        $lsb_release = find_lsb_release($lsb_release);
}
elsif($^O eq 'freebsd') {
	$www_dir  = '/usr/local/www/apache22/data/fci/';
	$owner    = 'root';
	$group    = 'www';
	$qd       = 'qd.freebsd';

	$apache_conf_dst = '/usr/local/etc/apache22/Includes/fci.conf';
}
elsif($^O eq 'openbsd') { 
	$www_dir  = '/var/apache2/htdocs/fci/';
	$owner    = 'root';
	$group    = '_apache2';
	$qd       = 'qd.openbsd';

	$apache_conf_dst = '/etc/apache2/extra/httpd-fci.conf';
}
elsif($^O eq 'cygwin') {
	$qd       = 'qd.exe';
	$ds       = '/';
}
elsif($^O eq 'MSWin32') {
	#Set default install directory
	if($ENV{PROGRAMFILES}) {
		$ENV{PROGRAMFILES} .= '\\';
	}
	else {
		$ENV{PROGRAMFILES} = 'C:\\Program Files\\';
	}
	
	$bin_dir  = $ENV{PROGRAMFILES}.'FCI\\'; 
	$conf_dir = $ENV{PROGRAMFILES}.'FCI\\'; 
	$data_dir = $ENV{PROGRAMFILES}.'FCI\\';
	$www_dir  = $ENV{PROGRAMFILES}.'FCI\\';
	$qd       = 'qd.exe';
}
elsif($^O eq 'darwin') {
	$owner    = 'root';
	$group    = 'www';
	$qd       = 'qd.darwin';
}


my %install_part = ( client => 0,
		     server => 0
		   );

my %cfg = ( 'self'               => 'install.pl',
            'version'            => '1.1',

	    'bin-dir'            => $bin_dir,
	    'conf-dir'           => $conf_dir,
	    'data-dir'           => $data_dir,
	    'www-dir'            => $www_dir,
	    
	    'fci-client-dir'     => $dir.'client'.$ds,
	    'fci-client-bin'     => $dir.'client'.$ds.'fci-client.pl',
	    'qdinfo-file'        => $dir.'client'.$ds.'qdinfo.dat',
	    'qd-bin'             => $dir.'client'.$ds.$qd,
	    'qd'                 => $qd,
	    
	    'fci-www-dir'        => $dir.'www'.$ds,
	    'fci-www-perms'      => $dir.'www'.$ds.'settings'.$ds.'site-data'.$ds.'permissions.xml',
	    'fci-www-owner'      => $owner,
	    'fci-www-group'      => $group,
	    'chmod-separator'    => ':',
	    
	    'fci-config-file'    => $dir.'www'.$ds.'settings'.$ds.'site-data'.$ds.'config.xml',

	    'lsb-release-bin'    => $lsb_release,

	    'fci-scripts-dir'    => $dir.'scripts'.$ds,
	    'fci-bin-dir'        => $dir.'bin'.$ds,
	
	    'upgrade'            => 0,
	    'force'              => 0,
	    'verbose'            => 1
	  );

# get command line parameters
Getopt::Mixed::init("bin-dir=s
		     conf-dir=s
		     data-dir=s
		     www-dir=s
		     owner=s
		     group=s
		     client
		     server
		     upgrade
		     force
                     verbose
                     help
		     b>bin-dir
		     n>conf-dir
		     d>data-dir
		     w>www-dir
		     o>owner
		     g>group
		     c>client
		     s>server
		     u>upgrade
		     f>force
		     v>verbose
		     h>help
                    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage();
		exit;
	}

	$cfg{'bin-dir'}       = $value if($option eq 'bin-dir');
	$cfg{'conf-dir'}      = $value if($option eq 'conf-dir');
	$cfg{'data-dir'}      = $value if($option eq 'data-dir');
	$cfg{'www-dir'}       = $value if($option eq 'www-dir');
	
	$cfg{'fci-www-owner'} = $value if($option eq 'owner');
	$cfg{'fci-www-group'} = $value if($option eq 'group');
	
	$cfg{'upgrade'}       = 1      if($option eq 'upgrade');
	$cfg{'force'}         = 1      if($option eq 'force');
	$cfg{'verbose'}       = 1      if($option eq 'verbose');

	$install_part{client} = 1 if($option eq 'client');
	$install_part{server} = 1 if($option eq 'server');
}
Getopt::Mixed::cleanup();

# suffix the directory separtator if it's missing
my @dir = ('bin-dir', 'conf-dir', 'data-dir', 'www-dir');
foreach(@dir) {
	$cfg{$_} .= $ds if(substr($cfg{$_}, -1, 1) ne $ds);
}

if(!$install_part{client} && !$install_part{server}) {
	print "Neither client nor server selected for installation, will not install anything!\n";
	print "Use --client and/or --server to install the respective parts.\n";
	exit;
}


# install the components
if($^O eq 'linux') {
	# Linux installation - Client
	# ---------------------------
	# copy client/qd            -> /usr/local/bin/qd
	# copy client/qdinfo.dat    -> /usr/local/share/fci/qdinfo.dat
	# copy client/fci-client.pl -> /usr/local/bin/fci-client.pl
	
	print "Host OS: Linux\n" if($cfg{verbose});

	# Detect Distro info via LSB
	if(-x $cfg{'lsb-release-bin'}) {
		my %lsb_release = get_lsb_release_data($cfg{'lsb-release-bin'});
			
		# Debian 3.1:   Debian
		# Kubuntu 6.10: Ubuntu
		if($lsb_release{distrib_id} ne '') {
			print "Distro : $lsb_release{distrib_id}\n" if($cfg{verbose});

			$cfg{distro} = $lsb_release{distrib_id};
		}
		else {
			print "Failed to detect Linux distribution using: $cfg{'lsb-release-bin'}\n";
		}
	}
	else {
		if(my $distro = detect_distro()) {
			print "Distro : $distro\n" if($cfg{verbose});

			$cfg{distro} = $distro;
		}
		else {
			print "Failed to detect Linux distribution!\n";
		}
	}
			
	if($cfg{distro} eq 'Debian' ||
	   $cfg{distro} eq 'Ubuntu' ||
	   $cfg{distro} eq 'Fedora' ||
	   $cfg{distro} eq 'CentOS' ||
	   $cfg{distro} eq 'Gentoo' ||
	   $cfg{distro} eq 'Arch'   ||
	   $cfg{distro} eq 'MandrivaLinux' ||
           $cfg{distro} eq 'Slackware' ||
	   $cfg{distro} eq 'MEPIS'
	) {
	   	# Known to work/Supported distro's
		print "FCI client & server have been tested on this distro.\n" if($cfg{verbose});

		if($cfg{distro} eq 'Fedora' ||
		   $cfg{distro} eq 'CentOS' ||
		   $cfg{distro} eq 'MandrivaLinux'
		) {
			$apache_conf_dst = '/etc/httpd/conf.d/fci.conf';
		}
		elsif($cfg{distro} eq 'Gentoo') {
			$apache_conf_dst = '/etc/apache2/vhosts.d/99_fci.conf';
		}
		elsif($cfg{distro} eq 'Arch') {
			$apache_conf_dst = '/etc/httpd/conf/extra/httpd-fci.conf';
		}
		elsif($cfg{distro} eq 'Slackware') {
			$apache_conf_dst = '/etc/httpd/extra/fci.conf';
		}
	}
	elsif($cfg{distro} eq 'SUSE LINUX') {
		# Only FCI client is known to work
		print "FCI client & server have been tested on this distro.\n" if($cfg{verbose});

		if($cfg{verbose}) {
			print "+ FCI client is supported on this distro.\n";
			print "- FCI server is not supported on this distro.\n";
			print "  See documentation/INSTALL.OpenSUSE for details.\n";
		}
	}
	else {
		# Support unverified for distro
		print "FCI has NOT been tested on this distro!\n" if($cfg{verbose});
	}
	print "\n" if($cfg{verbose});

	if($install_part{client} == 1) {
		# client files installation on Linux
		
		my $file_mode = 0444; # r--r--r--	
		
		# Slackware has no /usr/local/share by default
		if(!-e parent($cfg{'data-dir'})) {
			create_dir(parent($cfg{'data-dir'}), 0755);
		}

		# create the required directories
		my @dirs = qw/bin-dir data-dir/;
		foreach(@dirs) {
			if(!-e $cfg{$_}) {
				# permission 0755 = rxwr-xr-x
				
				create_dir($cfg{$_}, 0755);
			}
			if(!-w $cfg{$_}) {
				die "Cannot write directory: $cfg{$_}\nInstallation aborted!\n";
			}
		}
	
		if(!-e $cfg{'bin-dir'}.'qd' || $cfg{force} || $cfg{upgrade}) {
			# cp client/qd /usr/local/bin/qd
			
			print "Copying file $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd\n" if($cfg{verbose});
			
			copy($cfg{'qd-bin'}, $cfg{'bin-dir'}.'qd') || die "Cannot copy $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd ($!)\n";

			$file_mode = 0755; # rwxr-xr-x
			chmod $file_mode, $cfg{'bin-dir'}.$cfg{qd};
		}
		else {
			print "File already exists: $cfg{'bin-dir'}qd, skipping.\n" if($cfg{verbose});
		}
			
		if(!-e $cfg{'data-dir'}.'qdinfo.dat' || $cfg{force} || $cfg{upgrade}) {
			# cp client/qdinfo.dat /usr/local/share/fci/qdinfo.dat
			print "Copying file $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat\n" if($cfg{verbose});
			
			copy($cfg{'qdinfo-file'}, $cfg{'data-dir'}.'qdinfo.dat') || die "Cannot copy $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat ($!)\n";

			$file_mode = 0755; # rwxr-xr-x
			chmod $file_mode, $cfg{'data-dir'}.'qdinfo.dat';
		}
		else {
			print "File already exists: $cfg{'data-dir'}qdinfo.dat, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'fci-client.pl' || $cfg{force} || $cfg{upgrade}) {
			print "Copying file $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl\n" if($cfg{verbose});
		
			copy($cfg{'fci-client-bin'}, $cfg{'bin-dir'}.'fci-client.pl') || die "Cannot copy $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl ($!)\n";
			
			$file_mode = 0755; # rwxr-xr-x
			chmod $file_mode, $cfg{'bin-dir'}.'fci-client.pl';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}fci-client.pl, skipping.\n" if($cfg{verbose});
		}
	}
	
	# Linux installation - Server
	# ---------------------------
	# mkdir /var/www/fci
	# copy www/*     -> /var/www/fci/
	# copy scripts/* -> /usr/local/bin/
	# copy bin/*     -> /usr/local/bin/
	# chmod -R <user>:www-data /var/www/fci/
	
	if($install_part{server} == 1) {
		# server files installation on Linux
		
		# create the required directories
		
		# mkdir /usr/local/bin/ unless it already exists
		if(!-e $cfg{'bin-dir'}) {
			# permission 0755 = rxwr-xr-x
			
			create_dir($cfg{'bin-dir'}, 0755);
		}
		if(!-w $cfg{'bin-dir'}) {
			die "Cannot write directory: $cfg{'bin-dir'}\nInstallation aborted!\n";
		}
		
		# mkdir /var/www/fci/ unless it already exists
		if(!-e $cfg{'www-dir'}) {
			# permission 0775 = rxwrwxr-x
			
			create_dir($cfg{'www-dir'}, 0775);
			chmod 0775, $cfg{'www-dir'};
		}
		if(!-w $cfg{'www-dir'}) {
			die "Cannot write directory: $cfg{'www-dir'}\nInstallation aborted!\n";
		}

		
		# copy the www directory
		&copy_directory($cfg{'fci-www-dir'}, $cfg{'www-dir'}, '');
		print "\n" if($cfg{verbose});

		# create the symlinks in the target www directory;
		&create_symlinks($cfg{'www-dir'});
		print "\n" if($cfg{verbose});
	

		my $config = XMLin( $cfg{'fci-config-file'},
				    KeyAttr    => { application => 'name', site => 'name' },
				    ForceArray => [ 'application', 'site' ],
				    KeepRoot   => 1
				  ) || die $@;

		# copy the fci scripts
		my %script = ( 'fci-jmol-missing-projects.pl' => 'fci-jmol-missing-projects',
			       'fci-update-project-images.pl' => 'fci-update-project-images',
			       'fci-update-qd-files.pl'       => 'fci-update-qd-files',
			       'fci-update-stanford-files.pl' => 'fci-update-stanford-files',
			       'fci-update-xml-files.pl'      => 'fci-update-xml-files',
			       'fci-update-fahstats.pl'       => 'fci-update-fahstats',
			       'fci-update-eoc-stats.pl'      => 'fci-update-eoc-stats',
			       'fci-generate-queue-graphs.pl' => 'fci-generate-queue-graphs',
			       'fci-update-jmol-projects.pl'  => 'fci-update-jmol-projects',
		             );
			     
		opendir(DIR, $cfg{'fci-scripts-dir'}) || die "Cannot opendir: $cfg{'fci-scripts-dir'} ($!)";
		foreach(sort readdir DIR) {
			next if(/^\.{1,2}$/);

			my $file   = $_;
			my $source = $cfg{'fci-scripts-dir'}.$file;
			my $target = $cfg{'bin-dir'}.$file;

			if(!$script{$file}) {
				print "No support for $file, skipping.\n";
				next;
			}
			
			if(-f $source && (!-e $target || $cfg{force} || $cfg{upgrade})) {
				print "Copying script: $source -> $target\n" if($cfg{verbose});
				copy($source, $target);
				chmod 0755, $target;

				# save the path in the configuration file
				$config->{settings}->{application_list}->{application}->{$script{$file}}->{path}
					= $target;

			}
			else {
				if(-e $target) {
					print "File already exists: $target, skipping.\n" if($cfg{verbose});
					
					# save the path in the configuration file
					$config->{settings}->{application_list}->{application}->{$script{$file}}->{path}
						= $target;
				}
			}
		}
		closedir DIR;

		# if rasmol is already installed skip installing it, and use that one.
		#
		# On Debian (and derivatives), Gentoo, Slackware & Mandriva the 32 bit rasmol binary is actually:
		# /usr/lib/rasmol/rasmol.32

		my $found      = 0;
		my $check_path = '';

		if($cfg{distro} eq 'Debian' ||
		   $cfg{distro} eq 'Ubuntu' ||
		   $cfg{distro} eq 'MEPIS'  ||
		   $cfg{distro} eq 'Gentoo' ||
		   $cfg{distro} eq 'Slackware' ||
		   $cfg{distro} eq 'MandrivaLinux' 
		) {
			$check_path = '/usr/lib/rasmol/rasmol.32';
		}

		print "Path of: rasmol\n" if($cfg{verbose});
		if($check_path ne '') {
			print " $check_path\n" if($cfg{verbose});
			if(-e $check_path) {
				print "Found at: $check_path\n" if($cfg{verbose});
				$found = 1;
				$config->{settings}->{application_list}->{application}->{'rasmol'}->{path}
					= $check_path;
			}
		}
		
		if(!$found) {
			my @path = ('/usr/bin/', '/usr/local/bin/');
			foreach(@path) {
				print " ${_}rasmol\n" if($cfg{verbose});
				if(-e $_.'rasmol') {
					print "Found at: ${_}rasmol\n" if($cfg{verbose});
					$found = 1;
					$config->{settings}->{application_list}->{application}->{'rasmol'}->{path}
						= $_.'rasmol';
					last;
				}
			}
		}
	
		my @binary = qw/xyz2pdb/;
		if(!$found) {
			push @binary, 'rasmol';
		}
		
		foreach(@binary) {
			my $file   = $_;
			my $source = $cfg{'fci-bin-dir'}.$file;
			my $target = $cfg{'bin-dir'}.$file;
			
			#print "S: $source | T: $target\n";
			if(-f $source && (!-e $target || $cfg{force} || $cfg{upgrade})) {
				print "Copying executable: $source -> $target\n" if($cfg{verbose});
				copy($source, $target);
				chmod 0755, $target;

				# save the path in the configuration file
				$config->{settings}->{application_list}->{application}->{$file}->{path}
					= $target;
			}
			else {
				if(-e $target) {
					print "File already exists: $target, skipping.\n" if($cfg{verbose});

					# save the path in the configuration file
					$config->{settings}->{application_list}->{application}->{$file}->{path}
						= $target;
				}
			}
		}

		# find the paths to convert & htpasswd
		@binary = qw/convert htpasswd/;
		foreach(@binary) {
			my $file = $_;
			print "Path of: $file\n" if($cfg{verbose});
			
			my $found = 0;
			my @path = ('/usr/bin/', '/usr/sbin/', '/usr/local/bin/', '/usr/local/sbin/');
			foreach(@path) {
				print " ".$_.$file."\n";
				if(-e $_.$file) {
					print "Found at: $_$file\n" if($cfg{verbose});
					$found = 1;
					$config->{settings}->{application_list}->{application}->{$file}->{path}
						= $_.$file;
					last;
				}
			}
			if(!$found) {
				print "Error: Cannot find: $file\n";
				print "Please install the program and rerun this installer.\n";
				exit 1;
			}
		}
		print "\n" if($cfg{verbose});

		print "Saving application paths in configuration file.\n" if($cfg{verbose});
		XMLout( $config,
			OutputFile => $cfg{'fci-config-file'},
			KeyAttr    => { application => 'name', site => 'name' },
			NoAttr     => 1,
			KeepRoot   => 1
		      ) || die $@;
		print "\n" if($cfg{verbose});


		# allow the websever user the required permissions
		#
		# chmod -R <user>:www-data /var/www/fci/
		print "chown -R $cfg{'fci-www-owner'}$cfg{'chmod-separator'}$cfg{'fci-www-group'} $cfg{'www-dir'}\n";
		system('chown', '-R', $cfg{'fci-www-owner'}.$cfg{'chmod-separator'}.$cfg{'fci-www-group'}, $cfg{'www-dir'});

		# set the specific read, write and execute bits according to XML content
		my $permissions = XMLin( $cfg{'fci-www-perms'},
		                         KeyAttr    => { path => 'name' },
					 ForceArray => [ 'path' ],
					 KeepRoot   => 1
				       ) || die $@;

	        my $errors = 0;
	        foreach my $rel_path (sort { $a cmp $b } keys %{$permissions->{permissions}->{path}}) {
	                my $path = $cfg{'www-dir'} . $rel_path;
			
	                if($permissions->{permissions}->{path}->{$rel_path}->{read}) {
				print "chmod ug+r $path\n";
				system('chmod', 'ug+r', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{write}) {
				print "chmod ug+w $path\n";
				system('chmod', 'ug+w', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{execute}) {
				print "chmod ug+x $path\n";
				system('chmod', 'ug+x', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{symlink} && !-l $path) {
				print "Error: Not a symlink: $rel_path | $path\n";
	                }
	        }
		print "\n" if($cfg{verbose});

		print "Apache needs to be configured before the FCI server can accept client uploads.\n";
		print "An example configuration is supplied in: $apache_conf_src\n";

		if($cfg{'distro'} eq 'Debian' ||
		   $cfg{'distro'} eq 'Ubuntu') {
			print "This file can be copied to: $apache_conf_dst\n\n";

			print "This configuration can then be activated with:\n";
			print "a2ensite fci\n";
			print "/etc/init.d/apache2 reload\n\n";

			print "You may also need to enable the apreq module if you haven't done so already:\n";
			print "a2enmod apreq\n";
			print "/etc/init.d/apache2 force-reload\n\n";

			print "See documentation/INSTALL.Debian for more information.\n";
		}
		elsif($cfg{'distro'} eq 'Fedora' ||
		      $cfg{'distro'} eq 'CentOS' ||
		      $cfg{'distro'} eq 'MandrivaLinux') {
			print "This file can be copied to: $apache_conf_dst\n\n";

			print "This configuration can then be activated with:\n";
			print "service httpd reload\n\n";

			if($cfg{'distro'} eq 'Fedora') {
				print "See documentation/INSTALL.Fedora for more information.\n";
			}
			elsif($cfg{'distro'} eq 'CentOS') {
				print "See documentation/INSTALL.CentOS for more information.\n";
			}
			elsif($cfg{'distro'} eq 'MandrivaLinux') {
				print "See documentation/INSTALL.Mandriva for more information.\n";
			}
		}
		elsif($cfg{'distro'} eq 'Gentoo') {
			print "This file can be copied to: $apache_conf_dst\n\n";

			print "This configuration can then be activated with:\n";
			print "/etc/init.d/apache reload\n\n";

			print "See documentation/INSTALL.Gentoo for more information.\n";
		}
		elsif($cfg{'distro'} eq 'Arch') {
			print "This file can be copied to: $apache_conf_dst\n";
			print "And added to the apache configuration in: /etc/httpd/conf/httpd.conf\n";
			print "# FCI server\n";
			print "Include conf/extra/httpd-fci.conf\n\n";

			print "Also add the LoadModule statements for mod_perl and libapreq2 in: /etc/httpd/conf/httpd.conf\n";
			print "# Apache mod_perl\n";
			print "LoadModule perl_module modules/mod_perl.so\n";
			print "# libapreq2\n";
			print "LoadModule apreq_module modules/mod_apreq2.so\n\n";

			print "Be sure to modify the path of the FCI server files (/srv/http instead of /var/www/):\n";
			print "sed -e 's/var\\/www/srv\\/http/g' $apache_conf_dst > ${apache_conf_dst}.new\n";
			print "mv ${apache_conf_dst}.new $apache_conf_dst\n\n";

			print "This configuration can then be activated with:\n";
			print "/etc/rc.d/httpd restart\n\n";

			print "See documentation/INSTALL.Arch for more information.\n";
		}
		elsif($cfg{'distro'} eq 'Slackware') {
			print "This file can be copied to: $apache_conf_dst\n\n";

			print "This configuration can then be activated with:\n";
			print "/etc/rc.d/rc.httpd reload\n\n";

			print "See documentation/INSTALL.Slackware for more information.\n";
		}
		else {
			print "Please configure apache manually.\n\n";
			
			print "See documentation/INSTALL.Linux for more information.\n";
		}
	}
}
elsif($^O eq 'freebsd') {
	# FreeBSD installation - Client
	# -----------------------------
	# mkdir /usr/local/etc/fci/
	# copy client/qd            -> /usr/local/etc/fci/qd
	# copy client/qdinfo.dat    -> /usr/local/etc/fci/qdinfo.dat
	#
	# copy client/fci-client.pl -> /usr/local/bin/fci-client.pl
	
	print "Host OS: FreeBSD\n" if($cfg{verbose});
	if($install_part{client} == 1) {
		# client files installation on FreeBSD
	
		my $file_mode = 0444; # r--r--r--	
		
		if(!-e $cfg{'conf-dir'}) {
			# permission 0755 = rxwr-xr-x
			create_dir($cfg{'conf-dir'}, 0755);
		}
		if(!-w $cfg{'conf-dir'}) {
			die "Cannot write directory: $cfg{'conf-dir'}\nInstallation aborted!\n";
		}
		
		# /usr/local/bin/
		if(!-e $cfg{'bin-dir'}) {
			create_dir($cfg{'bin-dir'}, 0755);
		}
		if(!-w $cfg{'bin-dir'}) {
			die "Cannot write directory: $cfg{'bin-dir'}\nInstallation aborted!\n";
		}	
		
		# /usr/local/share/fci/
		if(!-e $cfg{'data-dir'}) {
			create_dir($cfg{'data-dir'}, 0755);
		}
		if(!-w $cfg{'data-dir'}) {
			die "Cannot write directory: $cfg{'data-dir'}\nInstallation aborted!\n";
		}	
		
		if(!-e $cfg{'bin-dir'}.'qd' || $cfg{force} || $cfg{upgrade}) {
			# cp client/qd.openbsd /usr/local/etc/fci/qd.openbsd
			print "Copying file $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd\n" if($cfg{verbose});
			
			copy($cfg{'qd-bin'}, $cfg{'bin-dir'}.'qd') || die "Cannot copy $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd ($!)\n";

			$file_mode = 0755; # rwxr-xr-x	
			chmod $file_mode, $cfg{'bin-dir'}.'qd';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}qd, skipping.\n" if($cfg{verbose});
		}
			
		if(!-e $cfg{'data-dir'}.'qdinfo.dat' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/qdinfo.dat /usr/local/etc/fci/qdinfo.dat
			print "Copying file $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat\n"	if($cfg{verbose});
			copy($cfg{'qdinfo-file'}, $cfg{'data-dir'}.'qdinfo.dat') || die "Cannot copy $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat ($!)\n";
			
			$file_mode = 0644; # rw-r--r--	
			chmod $file_mode, $cfg{'data-dir'}.'qdinfo.dat';
		}
		else {
			print "File already exists: $cfg{'data-dir'}qdinfo.dat, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'fci-client.pl' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/fci-client.pl /usr/local/bin/fci-client.pl
			print "Copying file $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl\n" if($cfg{verbose});
			
			$file_mode = 0755; # rwxr-xr-x
			
			copy($cfg{'fci-client-bin'}, $cfg{'bin-dir'}.'fci-client.pl') || die "Cannot copy $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl ($!)\n";
			
			chmod $file_mode, $cfg{'bin-dir'}.'fci-client.pl';
		}
	}
	
	# FreeBSD installation - Server
	# ---------------------------
	# mkdir /usr/local/www/apache22/fci
	# copy www/*     -> /usr/local/www/apache22/fci/
	# copy scripts/* -> /usr/local/bin/
	# copy bin/*     -> /usr/local/bin/
	# chmod -R <user>:www-data /usr/local/www/apache22/fci/
	
	if($install_part{server} == 1) {
		# server files installation on FreeBSD
		
		# create the required directories
		
		# mkdir /usr/local/bin/ unless it already exists
		if(!-e $cfg{'bin-dir'}) {
			# permission 0755 = rxwr-xr-x
			
			create_dir($cfg{'bin-dir'}, 0755);
		}
		if(!-w $cfg{'bin-dir'}) {
			die "Cannot write directory: $cfg{'bin-dir'}\nInstallation aborted!\n";
		}
		
		# mkdir /usr/local/www/apache22/fci/ unless it already exists
		if(!-e $cfg{'www-dir'}) {
			# permission 0775 = rxwrwxr-x
			
			create_dir($cfg{'www-dir'}, 0775);
			chmod 0775, $cfg{'www-dir'};
		}
		if(!-w $cfg{'www-dir'}) {
			die "Cannot write directory: $cfg{'www-dir'}\nInstallation aborted!\n";
		}

		
		# copy the www directory
		&copy_directory($cfg{'fci-www-dir'}, $cfg{'www-dir'}, '');
		print "\n" if($cfg{verbose});

		# create the symlinks in the target www directory;
		&create_symlinks($cfg{'www-dir'});
		print "\n" if($cfg{verbose});
	

		my $config = XMLin( $cfg{'fci-config-file'},
				    KeyAttr    => { application => 'name', site => 'name' },
				    ForceArray => [ 'application', 'site' ],
				    KeepRoot   => 1
				  ) || die $@;

		# copy the fci scripts
		my %script = ( 'fci-jmol-missing-projects.pl' => 'fci-jmol-missing-projects',
			       'fci-update-project-images.pl' => 'fci-update-project-images',
			       'fci-update-qd-files.pl'       => 'fci-update-qd-files',
			       'fci-update-stanford-files.pl' => 'fci-update-stanford-files',
			       'fci-update-xml-files.pl'      => 'fci-update-xml-files',
			       'fci-update-fahstats.pl'       => 'fci-update-fahstats',
			       'fci-update-eoc-stats.pl'      => 'fci-update-eoc-stats',
			       'fci-generate-queue-graphs.pl' => 'fci-generate-queue-graphs',
			       'fci-update-jmol-projects.pl'  => 'fci-update-jmol-projects',
		             );
			     
		opendir(DIR, $cfg{'fci-scripts-dir'}) || die "Cannot opendir: $cfg{'fci-scripts-dir'} ($!)";
		foreach(sort readdir DIR) {
			next if(/^\.{1,2}$/);

			my $file   = $_;
			my $source = $cfg{'fci-scripts-dir'}.$file;
			my $target = $cfg{'bin-dir'}.$file;

			if(!$script{$file}) {
				print "No support for $file, skipping.\n";
				next;
			}
			
			if(-f $source && (!-e $target || $cfg{'force'} || $cfg{'upgrade'})) {
				print "Copying script: $source -> $target\n" if($cfg{verbose});
				copy($source, $target);
				chmod 0755, $target;

				# save the path in the configuration file
				$config->{settings}->{application_list}->{application}->{$script{$file}}->{path}
					= $target;

			}
			else {
				if(-e $target) {
					print "File already exists: $target, skipping.\n" if($cfg{verbose});
					
					# save the path in the configuration file
					$config->{settings}->{application_list}->{application}->{$script{$file}}->{path}
						= $target;
				}
			}
		}
		closedir DIR;

		# If rasmol is already installed skip installing it, and use that one.
		#
		# On FreeBSD the 32 bit rasmol binary is actually:
		# /usr/local/bin/rasmol-32
		
                my $found      = 0;
		my $check_path = '/usr/local/bin/rasmol-32';

		if($check_path ne '') {
			if(-e $check_path) {
				print "Found at: $check_path\n" if($cfg{verbose});
				$found = 1;
				$config->{settings}->{application_list}->{application}->{'rasmol'}->{path}
					= $check_path;
			}
		}
		
		if(!$found) {
			my @path = ('/usr/bin/', '/usr/local/bin/');
			foreach(@path) {
				if(-e $_.'rasmol') {
					print "Found at: ${_}rasmol\n" if($cfg{verbose});
					$found = 1;
					$config->{settings}->{application_list}->{application}->{'rasmol'}->{path}
						= $_.'rasmol';
					last;
				}
			}
		}
		
		my @binary = qw/xyz2pdb/;
		if(!$found) {
			push @binary, 'rasmol';
		}
		
		foreach(@binary) {
			my $file   = $_;
			my $source = $cfg{'fci-bin-dir'}.$file;
			my $target = $cfg{'bin-dir'}.$file;
			
			#print "S: $source | T: $target\n";
			if(-f $source && (!-e $target || $cfg{force} || $cfg{'upgrade'})) {
				print "Copying executable: $source -> $target\n" if($cfg{verbose});
				copy($source, $target);
				chmod 0755, $target;

				# save the path in the configuration file
				$config->{settings}->{application_list}->{application}->{$file}->{path}
					= $target;
			}
			else {
				if(-e $target) {
					print "File already exists: $target, skipping.\n" if($cfg{verbose});

					# save the path in the configuration file
					$config->{settings}->{application_list}->{application}->{$file}->{path}
						= $target;
				}
			}
		}

		# find the paths to convert & htpasswd
		@binary = qw/convert htpasswd/;
		foreach(@binary) {
			my $file = $_;
			print "Path of: $file\n" if($cfg{verbose});
			
			my $found = 0;
			my @path = ('/usr/bin/', '/usr/sbin/', '/usr/local/bin/', '/usr/local/sbin/');
			foreach(@path) {
				print " ".$_.$file."\n";
				if(-e $_.$file) {
					print "Found at: $_$file\n" if($cfg{verbose});
					$found = 1;
					$config->{settings}->{application_list}->{application}->{$file}->{path}
						= $_.$file;
					last;
				}
			}
			if(!$found) {
				print "Error: Cannot find: $file\n";
				print "Please install the program and rerun this installer.\n";
				exit 1;
			}
		}
		print "\n" if($cfg{verbose});

		print "Saving application paths in configuration file.\n" if($cfg{verbose});
		XMLout( $config,
			OutputFile => $cfg{'fci-config-file'},
			KeyAttr    => { application => 'name', site => 'name' },
			NoAttr     => 1,
			KeepRoot   => 1
		      ) || die $@;
		print "\n" if($cfg{verbose});


		# allow the websever user the required permissions
		#
		# chmod -R <user>:www /usr/local/www/apache22/fci/
		print "chown -R $cfg{'fci-www-owner'}$cfg{'chmod-separator'}$cfg{'fci-www-group'} $cfg{'www-dir'}\n";
		system('chown', '-R', $cfg{'fci-www-owner'}.$cfg{'chmod-separator'}.$cfg{'fci-www-group'}, $cfg{'www-dir'});

		# set the specific read, write and execute bits according to XML content
		my $permissions = XMLin( $cfg{'fci-www-perms'},
		                         KeyAttr    => { path => 'name' },
					 ForceArray => [ 'path' ],
					 KeepRoot   => 1
				       ) || die $@;

	        my $errors = 0;
	        foreach my $rel_path (keys %{$permissions->{permissions}->{path}}) {
	                my $path = $cfg{'www-dir'} . $rel_path;
			
	                if($permissions->{permissions}->{path}->{$rel_path}->{read}) {
				print "chmod ug+r $path\n";
				system('chmod', 'ug+r', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{write}) {
				print "chmod ug+w $path\n";
				system('chmod', 'ug+w', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{execute}) {
				print "chmod ug+x $path\n";
				system('chmod', 'ug+x', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{symlink} && !-l $path) {
				print "Error: Not a symlink: $rel_path | $path\n";
	                }
	        }
		print "\n" if($cfg{verbose});

		print "Apache needs to be configured before the FCI server can accept client uploads.\n";
		print "An example configuration is supplied in: $apache_conf_src\n";

		print "This file can be copied to: $apache_conf_dst\n\n";
			
		print "This configuration can then be activated with:\n";
		print "/usr/local/sbin/apachectl restart\n\n";

		print "See documentation/INSTALL.FreeBSD for more information.\n";
	}
}
elsif($^O eq 'openbsd') {
	# OpenBSD installation - Client
	# -----------------------------
	# mkdir /usr/local/etc/fci/
	# copy client/qd            -> /usr/local/etc/fci/qd
	# copy client/qdinfo.dat    -> /usr/local/etc/fci/qdinfo.dat
	#
	# copy client/fci-client.pl -> /usr/local/bin/fci-client.pl
	
	print "Host OS: OpenBSD\n" if($cfg{verbose});
	if($install_part{client} == 1) {
		# client files installation on OpenBSD
	
		my $file_mode = 0444; # r--r--r--	
		
		# OpenBSD has no /usr/local/etc by default
		if(!-e parent($cfg{'conf-dir'})) {
			create_dir(parent($cfg{'conf-dir'}), 0755);
		}
			
		if(!-e $cfg{'conf-dir'}) {
			# permission 0755 = rxwr-xr-x
				
			create_dir($cfg{'conf-dir'}, 0755);
		}
		if(!-w $cfg{'conf-dir'}) {
			die "Cannot write directory: $cfg{'conf-dir'}\nInstallation aborted!\n";
		}
		
		# /usr/local/bin/
		if(!-e $cfg{'bin-dir'}) {
			create_dir($cfg{'bin-dir'}, 0755);
		}
		if(!-w $cfg{'bin-dir'}) {
			die "Cannot write directory: $cfg{'bin-dir'}\nInstallation aborted!\n";
		}	
		
		# /usr/local/share/fci/
		if(!-e $cfg{'data-dir'}) {
			create_dir($cfg{'data-dir'}, 0755);
		}
		if(!-w $cfg{'data-dir'}) {
			die "Cannot write directory: $cfg{'data-dir'}\nInstallation aborted!\n";
		}	
		
		if(!-e $cfg{'bin-dir'}.'qd' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/qd.openbsd /usr/local/etc/fci/qd.openbsd
			print "Copying file $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd\n" if($cfg{verbose});
			
			copy($cfg{'qd-bin'}, $cfg{'bin-dir'}.'qd') || die "Cannot copy $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd ($!)\n";

			$file_mode = 0755; # rwxr-xr-x	
			chmod $file_mode, $cfg{'bin-dir'}.'qd';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}qd, skipping.\n" if($cfg{verbose});
		}
			
		if(!-e $cfg{'data-dir'}.'qdinfo.dat' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/qdinfo.dat /usr/local/etc/fci/qdinfo.dat
			print "Copying file $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat\n"	if($cfg{verbose});
			copy($cfg{'qdinfo-file'}, $cfg{'data-dir'}.'qdinfo.dat') || die "Cannot copy $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat ($!)\n";
			
			$file_mode = 0644; # rw-r--r--	
			chmod $file_mode, $cfg{'data-dir'}.'qdinfo.dat';
		}
		else {
			print "File already exists: $cfg{'data-dir'}qdinfo.dat, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'fci-client.pl' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/fci-client.pl /usr/local/bin/fci-client.pl
			print "Copying file $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl\n" if($cfg{verbose});
			
			$file_mode = 0755; # rwxr-xr-x
			
			copy($cfg{'fci-client-bin'}, $cfg{'bin-dir'}.'fci-client.pl') || die "Cannot copy $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl ($!)\n";
			
			chmod $file_mode, $cfg{'bin-dir'}.'fci-client.pl';
		}
	}
	if($install_part{server} == 1) {
		# server files installation on OpenBSD

		warn "The server part of FCI is currently not supported on OpenBSD.\n";
	}
	
	# OpenBSD installation - Server
	# ---------------------------
	# mkdir /var/www/htdocs/fci
	# copy www/*     -> /var/www/htdocs/fci/
	# copy scripts/* -> /usr/local/bin/
	# copy bin/*     -> /usr/local/bin/
	# chmod -R <user>:www /var/www/htdocs/fci/
	
	if($install_part{server} == 1) {
		# server files installation on OpenBSD
		
		# create the required directories
		
		# mkdir /usr/local/bin/ unless it already exists
		if(!-e $cfg{'bin-dir'}) {
			# permission 0755 = rxwr-xr-x
			
			create_dir($cfg{'bin-dir'}, 0755);
		}
		if(!-w $cfg{'bin-dir'}) {
			die "Cannot write directory: $cfg{'bin-dir'}\nInstallation aborted!\n";
		}
		
		# mkdir /usr/local/www/apache22/fci/ unless it already exists
		if(!-e $cfg{'www-dir'}) {
			# permission 0775 = rxwrwxr-x
			
			create_dir($cfg{'www-dir'}, 0775);
			chmod 0775, $cfg{'www-dir'};
		}
		if(!-w $cfg{'www-dir'}) {
			die "Cannot write directory: $cfg{'www-dir'}\nInstallation aborted!\n";
		}

		
		# copy the www directory
		&copy_directory($cfg{'fci-www-dir'}, $cfg{'www-dir'}, '');
		print "\n" if($cfg{verbose});

		# create the symlinks in the target www directory;
		&create_symlinks($cfg{'www-dir'});
		print "\n" if($cfg{verbose});
	

		my $config = XMLin( $cfg{'fci-config-file'},
				    KeyAttr    => { application => 'name', site => 'name' },
				    ForceArray => [ 'application', 'site' ],
				    KeepRoot   => 1
				  ) || die $@;

		# copy the fci scripts
		my %script = ( 'fci-jmol-missing-projects.pl' => 'fci-jmol-missing-projects',
			       'fci-update-project-images.pl' => 'fci-update-project-images',
			       'fci-update-qd-files.pl'       => 'fci-update-qd-files',
			       'fci-update-stanford-files.pl' => 'fci-update-stanford-files',
			       'fci-update-xml-files.pl'      => 'fci-update-xml-files',
			       'fci-update-fahstats.pl'       => 'fci-update-fahstats',
			       'fci-update-eoc-stats.pl'      => 'fci-update-eoc-stats',
			       'fci-generate-queue-graphs.pl' => 'fci-generate-queue-graphs',
			       'fci-update-jmol-projects.pl'  => 'fci-update-jmol-projects',
		             );
			     
		opendir(DIR, $cfg{'fci-scripts-dir'}) || die "Cannot opendir: $cfg{'fci-scripts-dir'} ($!)";
		foreach(sort readdir DIR) {
			next if(/^\.{1,2}$/);

			my $file   = $_;
			my $source = $cfg{'fci-scripts-dir'}.$file;
			my $target = $cfg{'bin-dir'}.$file;

			if(!$script{$file}) {
				print "No support for $file, skipping.\n";
				next;
			}
			
			if(-f $source && (!-e $target || $cfg{'force'} || $cfg{'upgrade'})) {
				print "Copying script: $source -> $target\n" if($cfg{verbose});
				copy($source, $target);
				chmod 0755, $target;

				# save the path in the configuration file
				$config->{settings}->{application_list}->{application}->{$script{$file}}->{path}
					= $target;

			}
			else {
				if(-e $target) {
					print "File already exists: $target, skipping.\n" if($cfg{verbose});
					
					# save the path in the configuration file
					$config->{settings}->{application_list}->{application}->{$script{$file}}->{path}
						= $target;
				}
			}
		}
		closedir DIR;

		# if rasmol is already installed
		# skip installing it, and use that one
                my $found = 0;
		my @path = ('/usr/bin/', '/usr/local/bin/');
		foreach(@path) {
			if(-e $_.'rasmol') {
				print "Found at: ${_}rasmol\n" if($cfg{verbose});
				$found = 1;
				$config->{settings}->{application_list}->{application}->{'rasmol'}->{path}
					= $_.'rasmol';
				last;
			}
		}
		
		my @binary = qw/xyz2pdb/;
		if(!$found) {
			push @binary, 'rasmol';
		}
		
		foreach(@binary) {
			my $file   = $_;
			my $source = $cfg{'fci-bin-dir'}.$file;
			my $target = $cfg{'bin-dir'}.$file;
			
			#print "S: $source | T: $target\n";
			if(-f $source && (!-e $target || $cfg{force} || $cfg{'upgrade'})) {
				print "Copying executable: $source -> $target\n" if($cfg{verbose});
				copy($source, $target);
				chmod 0755, $target;

				# save the path in the configuration file
				$config->{settings}->{application_list}->{application}->{$file}->{path}
					= $target;
			}
			else {
				if(-e $target) {
					print "File already exists: $target, skipping.\n" if($cfg{verbose});

					# save the path in the configuration file
					$config->{settings}->{application_list}->{application}->{$file}->{path}
						= $target;
				}
			}
		}

		# find the paths to convert & htpasswd
		@binary = qw/convert htpasswd/;
		foreach(@binary) {
			my $file = $_;
			print "Path of: $file\n" if($cfg{verbose});
			
			my $found = 0;
			my @path = ('/usr/bin/', '/usr/sbin/', '/usr/local/bin/', '/usr/local/sbin/');
			foreach(@path) {
				print " ".$_.$file."\n";
				if(-e $_.$file) {
					print "Found at: $_$file\n" if($cfg{verbose});
					$found = 1;
					$config->{settings}->{application_list}->{application}->{$file}->{path}
						= $_.$file;
					last;
				}
			}
			if(!$found) {
				print "Error: Cannot find: $file\n";
				print "Please install the program and rerun this installer.\n";
				exit 1;
			}
		}
		print "\n" if($cfg{verbose});

		print "Saving application paths in configuration file.\n" if($cfg{verbose});
		XMLout( $config,
			OutputFile => $cfg{'fci-config-file'},
			KeyAttr    => { application => 'name', site => 'name' },
			NoAttr     => 1,
			KeepRoot   => 1
		      ) || die $@;
		print "\n" if($cfg{verbose});


		# allow the websever user the required permissions
		#
		# chmod -R <user>:www /usr/local/www/apache22/fci/
		print "chown -R $cfg{'fci-www-owner'}$cfg{'chmod-separator'}$cfg{'fci-www-group'} $cfg{'www-dir'}\n";
		system('chown', '-R', $cfg{'fci-www-owner'}.$cfg{'chmod-separator'}.$cfg{'fci-www-group'}, $cfg{'www-dir'});

		# set the specific read, write and execute bits according to XML content
		my $permissions = XMLin( $cfg{'fci-www-perms'},
		                         KeyAttr    => { path => 'name' },
					 ForceArray => [ 'path' ],
					 KeepRoot   => 1
				       ) || die $@;

	        my $errors = 0;
	        foreach my $rel_path (keys %{$permissions->{permissions}->{path}}) {
	                my $path = $cfg{'www-dir'} . $rel_path;
			
	                if($permissions->{permissions}->{path}->{$rel_path}->{read}) {
				print "chmod ug+r $path\n";
				system('chmod', 'ug+r', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{write}) {
				print "chmod ug+w $path\n";
				system('chmod', 'ug+w', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{execute}) {
				print "chmod ug+x $path\n";
				system('chmod', 'ug+x', $path);
	                }
	                if($permissions->{permissions}->{path}->{$rel_path}->{symlink} && !-l $path) {
				print "Error: Not a symlink: $rel_path | $path\n";
	                }
	        }
		print "\n" if($cfg{verbose});

		print "Apache needs to be configured before the FCI server can accept client uploads.\n";
		print "An example configuration is supplied in: $apache_conf_src\n";

		print "This file can be copied to: $apache_conf_dst\n\n";
			
		print "This configuration can then be activated with:\n";
		print "/usr/local/sbin/apachectl2 restart\n\n";

		print "See documentation/INSTALL.OpenBSD for more information.\n";
	}
}
elsif($^O eq 'cygwin') {
	# Cygwin (on Windows) installation - Client
	# -----------------------------------------
	# copy client/qd.exe        -> /usr/local/bin/qd.exe
	# copy client/qdinfo.dat    -> /usr/local/share/fci/qdinfo.dat
	# copy client/fci-client.pl -> /usr/local/bin/fci-client.pl
	
	print "Host OS: Cygwin (on Windows)\n" if($cfg{verbose});
	if($install_part{client} == 1) {
		# client files installation on Linux
		
		my $file_mode = 0444; # r--r--r--	
		
		# create the required directories
		my @dirs = qw/bin-dir data-dir/;
		foreach(@dirs) {
			if(!-e $cfg{$_}) {
				# Cygwin has no /usr/local/share by default
				if(!-e parent($cfg{$_})) {
					create_dir(parent($cfg{$_}), 0755);
				}

				# permission 0755 = rxwr-xr-x
				create_dir($cfg{$_}, 0755);
			}
			if(!-w $cfg{$_}) {
				die "Cannot write directory: $cfg{$_}\nInstallation aborted!\n";
			}
		}
	
		if(!-e $cfg{'bin-dir'}.'qd.exe' || $cfg{force} || $cfg{'upgrade'}) {
			# cp client/qd /usr/local/bin/qd.exe
			
			print "Copying file $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd.exe\n" if($cfg{verbose});
			
			copy($cfg{'qd-bin'}, $cfg{'bin-dir'}.'qd.exe') || die "Cannot copy $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd.exe ($!)\n";

			$file_mode = 0755; # rwxr-xr-x
			chmod $file_mode, $cfg{'bin-dir'}.'qd.exe';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}qd.exe, skipping.\n" if($cfg{verbose});
		}
			
		if(!-e $cfg{'data-dir'}.'qdinfo.dat' || $cfg{force} || $cfg{'upgrade'}) {
			# cp client/qdinfo.dat /usr/local/share/fci/qdinfo.dat
			print "Copying file $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat\n" if($cfg{verbose});
			
			copy($cfg{'qdinfo-file'}, $cfg{'data-dir'}.'qdinfo.dat') || die "Cannot copy $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat ($!)\n";

			$file_mode = 0755; # rwxr-xr-x
			chmod $file_mode, $cfg{'data-dir'}.'qdinfo.dat';
		}
		else {
			print "File already exists: $cfg{'data-dir'}qdinfo.dat, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'fci-client.pl' || $cfg{force} || $cfg{'upgrade'}) {
			print "Copying file $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl\n" if($cfg{verbose});
		
			copy($cfg{'fci-client-bin'}, $cfg{'bin-dir'}.'fci-client.pl') || die "Cannot copy $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl ($!)\n";
			
			$file_mode = 0755; # rwxr-xr-x
			chmod $file_mode, $cfg{'bin-dir'}.'fci-client.pl';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}fci-client.pl, skipping.\n" if($cfg{verbose});
		}
	}
	
}
elsif($^O eq 'MSWin32') {
	# Windows installation
	# 
	# Client
	# ------
	# mkdir C:\Program Files\fci\
	# copy client\qd.exe        -> C:\Program Files\FCI\qd.exe
	# copy client\qdinfo.dat    -> C:\Program Files\FCI\qdinfo.dat
	#
	# copy client\fci-client.pl -> C:\Program Files\FCI\fci-client.pl
	
	$cfg{'qd-bin'}         = $dir.'client\\qd.exe';
	$cfg{'qd'}             = 'qd.exe';
	
	print "Host OS: Windows\n" if($cfg{verbose});
	if($install_part{client} == 1) {
		# client files installation on Windows
		
		my $dir_mode = 0777; # rwxrwxrwx
		
		# create the required directories
		if(!-e $cfg{'bin-dir'}) {
			mkdir $cfg{'bin-dir'}, $dir_mode || die "Cannot create directory: $cfg{'bin-dir'} ($!)";
		}
		
		my $file_mode = 0777; # rwxrwxrwx
		if(!-e $cfg{'bin-dir'}.$cfg{qd} || $cfg{'force'} || $cfg{'upgrade'}) {
			# copy client\qd.exe C:\Program Files\fci\qd.exe

			print "Copying file $cfg{'qd-bin'} -> $cfg{'bin-dir'}$cfg{qd}\n" if($cfg{verbose});
			
			copy($cfg{'qd-bin'}, $cfg{'bin-dir'}.$cfg{qd}) || die "Cannot copy $cfg{'qd-bin'} -> $cfg{'bin-dir'}$cfg{qd} ($!)\n";
			chmod $file_mode, $cfg{'bin-dir'}.$cfg{qd};
		}
		else {
			print "File already exists: $cfg{'bin-dir'}$cfg{qd}, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'qdinfo.dat' || $cfg{'force'} || $cfg{'upgrade'}) {
			# copy client\qdinfo.dat C:\Program Files\fci\qdinfo.dat
			print "Copying file $cfg{'qdinfo-file'} -> $cfg{'bin-dir'}qdinfo.dat\n" if($cfg{verbose});
			
			copy($cfg{'qdinfo-file'}, $cfg{'bin-dir'}.'qdinfo.dat') || die "Cannot copy $cfg{'qdinfo-file'} -> $cfg{'conf-dir'}qdinfo.dat ($!)\n";

			chmod $file_mode, $cfg{'bin-dir'}.'qdinfo.dat';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}qdinfo.dat, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'fci-client.pl' || $cfg{'force'} || $cfg{'upgrade'}) {
			# copy client\fci-client.pl C:\Program Files\fci\fci-client.pl
			
			print "Copying file $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl\n" if($cfg{verbose});
			
			$file_mode = 0777; # rwxrwxrwx
	
			copy($cfg{'fci-client-bin'}, $cfg{'bin-dir'}.'fci-client.pl') || die "Cannot copy $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl ($!)\n";
		
			chmod $file_mode, $cfg{'bin-dir'}.'fci-client.pl';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}fci-client.pl, skipping.\n" if($cfg{verbose});
		}
	}
	if($install_part{server} == 1) {
		# server files installation on Windows

		warn "\nThe server part of FCI is not supported on Windows!\n\n";
	}
}
elsif($^O eq 'darwin') {
	# Mac OS X installation - Client
	# ------------------------------
	# mkdir /usr/local/etc/fci/
	# copy client/qd            -> /usr/local/etc/fci/qd
	# copy client/qdinfo.dat    -> /usr/local/etc/fci/qdinfo.dat
	#
	# copy client/fci-client.pl -> /usr/local/bin/fci-client.pl
	
	print "Host OS: Mac OS X\n" if($cfg{verbose});
	if($install_part{client} == 1) {
		# client files installation on Mac OS X
	
		my $file_mode = 0444; # r--r--r--	
		
		if(!-e $cfg{'conf-dir'}) {
			# permission 0755 = rxwr-xr-x
				
			create_dir($cfg{'conf-dir'}, 0755);
		}
		if(!-w $cfg{'conf-dir'}) {
			die "Cannot write directory: $cfg{'conf-dir'}\nInstallation aborted!\n";
		}
		
		# /usr/local/bin/
		if(!-e $cfg{'bin-dir'}) {
			create_dir($cfg{'bin-dir'}, 0755);
		}
		if(!-w $cfg{'bin-dir'}) {
			die "Cannot write directory: $cfg{'bin-dir'}\nInstallation aborted!\n";
		}	
		
		# /usr/local/share/fci/
		if(!-e $cfg{'data-dir'}) {
			create_dir($cfg{'data-dir'}, 0755);
		}
		if(!-w $cfg{'data-dir'}) {
			die "Cannot write directory: $cfg{'data-dir'}\nInstallation aborted!\n";
		}	
		
		if(!-e $cfg{'bin-dir'}.'qd' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/qd.darwin /usr/local/bin/qd
			print "Copying file $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd\n" if($cfg{verbose});
			
			copy($cfg{'qd-bin'}, $cfg{'bin-dir'}.'qd') || die "Cannot copy $cfg{'qd-bin'} -> $cfg{'bin-dir'}qd ($!)\n";

			$file_mode = 0755; # rwxr-xr-x	
			chmod $file_mode, $cfg{'bin-dir'}.'qd';
		}
		else {
			print "File already exists: $cfg{'bin-dir'}qd, skipping.\n" if($cfg{verbose});
		}
			
		if(!-e $cfg{'data-dir'}.'qdinfo.dat' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/qdinfo.dat /usr/local/etc/fci/qdinfo.dat
			print "Copying file $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat\n"	if($cfg{verbose});
			copy($cfg{'qdinfo-file'}, $cfg{'data-dir'}.'qdinfo.dat') || die "Cannot copy $cfg{'qdinfo-file'} -> $cfg{'data-dir'}qdinfo.dat ($!)\n";
			
			$file_mode = 0644; # rw-r--r--	
			chmod $file_mode, $cfg{'data-dir'}.'qdinfo.dat';
		}
		else {
			print "File already exists: $cfg{'data-dir'}qdinfo.dat, skipping.\n" if($cfg{verbose});
		}
		
		if(!-e $cfg{'bin-dir'}.'fci-client.pl' || $cfg{'force'} || $cfg{'upgrade'}) {
			# cp client/fci-client.pl /usr/local/bin/fci-client.pl
			print "Copying file $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl\n" if($cfg{verbose});
			
			$file_mode = 0755; # rwxr-xr-x
			
			copy($cfg{'fci-client-bin'}, $cfg{'bin-dir'}.'fci-client.pl') || die "Cannot copy $cfg{'fci-client-bin'} -> $cfg{'bin-dir'}fci-client.pl ($!)\n";
			
			chmod $file_mode, $cfg{'bin-dir'}.'fci-client.pl';
		}
	}
	if($install_part{server} == 1) {
		# server files installation on Mac OS X

		warn "The server part of FCI is currently not supported on Mac OS X.\n";
	}
}

################################[ Subroutines ]################################

sub create_symlinks {
	my $www_dir      = shift;
	my $settings_dir = $www_dir.'settings/';
	my $old_dir      = Cwd::cwd;

	print "Creating symlinks.\n" if($cfg{verbose});
	
	print "Changing to directory: $www_dir\n";
	Cwd::chdir($www_dir);
	
	# /var/www/fci/qd-files/                -> /var/www/fci/qd-data/
	if(!-e 'qd-files') {
		print "Linking qd-data/ -> qd-files\n" if($cfg{verbose});
		symlink('qd-data/', 'qd-files');
	}
	else {
		print "File already exists: qd-files, skipping.\n" if($cfg{verbose});
	}
	
	print "Changing to directory: $settings_dir\n";
	Cwd::chdir($settings_dir);

	# /var/www/fci/settings/index.pl        -> /var/www/fci/index.pl
	if(!-e 'index.pl') {
		print "Linking ../index.pl -> index.pl\n" if($cfg{verbose});
		symlink('../index.pl', 'index.pl');
	}
	else {
		print "File already exists: index.pl, skipping.\n" if($cfg{verbose});
	}

	# /var/www/fci/settings/permissions.plc -> /var/www/fci/permissions.plc
	if(!-e 'permissions.plc') {
		print "Linking ../permissions.plc -> permissions.plc\n" if($cfg{verbose});
		symlink('../permissions.plc', 'permissions.plc');
	}
	else {
		print "File already exists: permissions.plc, skipping.\n" if($cfg{verbose});
	}
	
	# /var/www/fci/settings/stylesheet.css  -> /var/www/fci/stylesheet.css
	if(!-e 'stylesheet.css') {
		print "Linking ../stylesheet.css -> stylesheet.css\n" if($cfg{verbose});
		symlink('../stylesheet.css', 'stylesheet.css');
	}
	else {
		print "File already exists: stylesheet.css, skipping.\n" if($cfg{verbose});
	}

	print "Changing back to directory: $old_dir\n";
	Cwd::chdir($old_dir);
}

sub copy_directory {
	my $source_dir = shift;
	my $target_dir = shift;
	my $indent     = shift;

	print $indent." Copying directory: $source_dir -> $target_dir\n" if($cfg{verbose});
	if(!-r $source_dir) {
		print "Error: Cannot read directory: $source_dir\n";
		exit 1;
	}
	if(!-w $target_dir) {
		print "Error: Cannot write directory: $target_dir\n";
		exit 1;
	}

	opendir(my $dh, $source_dir) || die "Error: Cannot opendir: $source_dir ($!)";
	foreach(sort readdir $dh) {
		next if(/^\.{1,2}$/);

		my $source = $source_dir.$_;
		my $target = $target_dir.$_;
		if(-l $source) {
			#print "$indent l: $_\n" if($cfg{verbose});
			#symlink($source, $target);
		}
		elsif(-d $source) {
			print "$indent d: $_\n" if($cfg{verbose});
			
			$source .= '/';
			$target .= '/';
			
			mkdir $target, $dir_mode;

			&copy_directory($source, $target, "$indent ");
		}
		elsif(-f $source) {
			print "$indent f: $_\n" if($cfg{verbose});

			# We don't want to overwrite existing files in the qd-data, xml-data & site-data directories.
			# These contain the data of the FCI server which must be preserved, unless forced.
			#
			# The only exception to this rule is site-data/permissions.xml, which will have keys
			# that exist only in the source copied to the target.
			# The whole file is overwritten when forced.
			#
			# We also don't want to overwrite the existing files in the stanford-data directory.
			# This contains the data files dowloaded from the Stanford servers, whereas FCI 
			# contains empty files as placeholders for the stanford-data files. 
			# These files must never be overwritten, not even when forced.

			if(-e $target && $target =~ /\/site-data\/permissions.xml/ && !$cfg{force}) {
				my $source_xml = XMLin( $source,
						        KeyAttr    => { path => 'name' },
							ForceArray => [ 'path' ],
							KeepRoot   => 1
					              ) || die $@;

				my $target_xml = XMLin( $target,
						        KeyAttr    => { path => 'name' },
							ForceArray => [ 'path' ],
							KeepRoot   => 1
					              ) || die $@;

				my $modified = 0;
				foreach my $path (keys %{$source_xml->{permissions}->{path}}) {
					if(!exists $target_xml->{permissions}->{path}->{$path}) {
						$target_xml->{permissions}->{path}->{$path}
							= $source_xml->{permissions}->{path}->{$path};

						$modified = 1;
					}
				}

				if($modified) {
					print "$indent  File already exists: $target, merged.\n";

					XMLout( $target_xml,
						OutputFile => $target,
						KeyAttr    => { path => 'name' },
						NoAttr     => 1,
						KeepRoot   => 1
					      ) || die $@;
				}
				else {
					print "$indent  File already exists: $target, skipped.\n";
				}

				next;
			}
			elsif(-e $target && $target =~ /\/(qd|xml|site)-data\// && !$cfg{force}) {
				print "$indent  File already exists: $target, skipping.\n";
				next;
			}
			elsif(-e $target && $target =~ /\/stanford-data\//) {
				print "$indent  File already exists: $target, skipping.\n";
				next;
			}


			if(!-e $target || $cfg{force} || $cfg{upgrade}) {
				print "$indent  Copying to: $target\n";
				copy($source, $target);
			}
			else {
				print "$indent  File already exists: $target, skipping.\n";
			}
		}
		else {
			print "$indent u: $_\n" if($cfg{verbose});
			print "Error: No handling for: $source\n";
			closedir $dh;
			exit 1;
		}
	}
	closedir $dh;

	return;
}

sub create_dir {
	my $dir  = $_[0];
	my $mode = $_[1];
	
	print "Creating directory $dir\n" if($cfg{verbose});
	my $parent = parent($dir);
	if(-w $parent) {
		mkdir($dir, $mode) || die "Cannot mkdir $dir, $mode! ($!)\n";
	}
	else {
		die "Cannot create directory: $dir (Cannot write $parent)\n";
	}
}

sub parent {
	my $dir = $_[0];
		
	my $parent = $dir;
	# Only Windows uses \ as directory delimiter
	# Linux and the BSDs use /
	if($^O ne 'MSWin32') {
		$parent =~ s/^(.*\/).+?$/$1/;
	}
	else {
		$parent =~ s/^(.*\\).+?$/$1/;
	}
	print "Parent of $dir is $parent\n" if($cfg{verbose});
	
	return $parent;
}

##############
#
#     Routine: find_lsb_release()
#  Parameters: default return value
#     Returns: path to lsb_release binary
# Description: find the path to the lsb_release binary or return the default value;
# 

sub find_lsb_release {
	my $file = shift;

	if(chomp(my $path = `which lsb_release 2> /dev/null`)) {
		return $path;
	}
	else {
		my @dir = qw|/bin/ /usr/bin/ /usr/local/bin/|;

		foreach(@dir) {
			my $bin  = $_;
			   $bin .= '/' if(substr($bin, -1, 1) ne '/');
			   $bin .= 'lsb_release';

			return $bin if(-x $bin);
		}
	}

	return $file;
}


##############
#
#     Routine: get_lsb_release_data()
#  Parameters: lsb_release path
#     Returns: hash with LSB info
# Description: get the Linux Standard Base info for this host
# 

sub get_lsb_release_data {
	my $file = shift;
	
	my %info;
	if(-x $file) {
		#$ lsb_release -a
		#LSB Version:    n/a
		#Distributor ID: Debian
		#Description:    Debian GNU/Linux
		#Release:        3.1
		#Codename:       sarge
	
		if(my $data = `$file --all 2>&1`) {
			if($data =~ /LSB Version:\t(.*)/) {
				$info{lsb_version} = $1;
			}
			if($data =~ /Distributor ID:\t(.*)/) {
				$info{distrib_id} = $1;
			}
			if($data =~ /Description:\t(.*)/) {
				$info{distrib_description} = $1;
			}
			if($data =~ /Release:\t(.*)/) {
				$info{distrib_release} = $1;
			}
			if($data =~ /Codename:\t(.*)/) {
				$info{distrib_codename} = $1;
			}
		}
	}

	return %info;
}

sub detect_distro {
	# Detect which Linux distribution is being used,
	# by checking which /etc/*release files are available

	my $distro_name = '';

	if(-r "/etc/debian_version") {
		# Debian, Ubuntu or other derivative
		$distro_name = "Debian";

		# /etc/debian_version
		# 3.1
	}
	elsif(-r "/etc/gentoo-release") {
		$distro_name = "Gentoo";

		# /etc/gentoo-release
		# Gentoo Base System version 1.12.1
	}
	elsif(-r "/etc/mandrake-release" || -r "/etc/mandrakelinux-release" || -r "/etc/mandriva-release") {
		$distro_name = "MandrivaLinux";

		# /etc/mandrake-release
		# Mandrake Linux release 8.1 (Vitamin) for i586

		# /etc/mandriva-release
		# Mandriva Linux release 2009.0 (Official) for i586
	}
	elsif(-r "/etc/fedora-release") {
		$distro_name = "Fedora";

		# /etc/fedora-release
		# Fedora Core release 6 (Zod)
	}
	elsif(-r "/etc/redhat-release") {
		# Red Hat, or derivative

		$distro_name = "Red Hat";

		# /etc/redhat-release
		# Red Hat Enterprise Linux AS release 4 (Nahant Update 3)
		# CentOS release 5.2 (Final)
		
		open(F, "/etc/redhat-release") || die "Errror: Cannot open file ($!)";
		while(<F>) {
			if(/^CentOS/) {
				$distro_name = "CentOS";
				last;
			}
		}
		close F;
	}
	elsif(-r "/etc/SuSE-release") {
		$distro_name = "SUSE LINUX";

		# /etc/SuSE-release
		# openSUSE 10.2 (i586)
		# VERSION = 10.2
	}
	elsif(-r "/etc/arch-release") {
		$distro_name = "Arch";

		# /etc/arch-release
		# <empty>
	}
	elsif(-r "/etc/slackware-version" || -r "/etc/slackware-release") {
		# Slackware, or derivative

		$distro_name = "Slackware";
		
		# /etc/slackware-version
		# /etc/slackware-release
		# Slackware release 8.1 (Midas)

		if(-r "/etc/slax-version") {
			# SLAX is a derivative of Slackware
			
			$distro_name = "SLAX";
			
			# /etc/slax-version
			# SLAX 5.1.8
		}
	}
	elsif(-r "/etc/UnitedLinux-release") {
		$distro_name = "UnitedLinux";

		# /etc/UnitedLinux-release
		# UnitedLinux 1.0 (ia64)
		# VERSION = 1.0
	}
	elsif(-r "/etc/sun-release") {
		$distro_name = "Sun";

		# /etc/sun-release
		# Sun Java Desktop System, Release 2 -build 10b (GA)
		# Assembled 30 March 2004
	}
	elsif(-r "/etc/yellowdog-release") {
		$distro_name = "Yellow Dog Linux";

		# /etc/yellowdog-release
		# Yellow Dog Linux release 4.0 (Orion)
	}

	return $distro_name;
}

sub print_usage {
	print "Folding\@Home Client Info Installer v$cfg{version}\n\n";
	print "Usage: install.pl [--client|--server] [OPTIONS]\n\n";
	print "Options:\n";
	print "-c, --client           Install the client component\n";
	print "-s, --server           Install the server component\n";
	print "\n";
	if($^O eq 'MSWin32') {
		print "-b, --bin-dir <PATH>   Path to install the files         ($cfg{'bin-dir'})\n";
	}
	else {
		print "-b, --bin-dir <PATH>   Path of executable files          ($cfg{'bin-dir'})\n";
		print "-n, --conf-dir <PATH>  Path of configuration files       ($cfg{'conf-dir'})\n";
		print "-d, --data-dir <PATH>  Path of the data files            ($cfg{'data-dir'})\n";
		print "-w, --www-dir <PATH>   Path of the website files         ($cfg{'www-dir'})\n";
		print "\n";
		print "-o, --owner <NAME>     Username to own the website files ($cfg{'fci-www-owner'})\n";
		print "-g, --group <NAME>     Username to own the website files ($cfg{'fci-www-group'})\n";
	}
	print "\n";
	print "-u, --upgrade          Upgrade already installed files, except local configuration\n";
	print "-f, --force            Install files & directories even if they already exist\n";
	print "-v, --verbose          Enable verbose output\n";
	print "-h, --help             Print this usage information\n";
}
