#!/usr/bin/perl -w
# 
# Generate images from a .xyz or .pdb file
# 

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use LWP::UserAgent;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";
use Image::Magick;

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my $www_data_dir     = '/var/www/fci/';
my $settings_dir     = 'settings/';
my $project_data_dir = 'project-data/';
my $xml_data_dir     = 'xml-data/';
my $site_data_dir    = $settings_dir.'site-data/';

my %cfg = ( self                     => 'fci-update-project-images.pl',
            version                  => '0.1.3',
	    
	    verbose                  => 0,
	    quiet                    => 0,
	    
	    'www-data-dir'           => $www_data_dir,
	    'settings-dir'           => $settings_dir,
	    'project-data-dir'       => $project_data_dir,
	    'site-data-dir'          => $site_data_dir,
	    'site-config-file'       => $site_data_dir.'config.xml',
	    'xml-data-dir'           => $xml_data_dir,
	    'known-projects-file'    => $xml_data_dir.'known-projects.xml',
	    'project-images-file'    => $xml_data_dir.'project-images.xml',

	    'project'                => 0,
	    'image-type'	     => '',
	    
	    'action'                 => '',
	    'generate-pdb'           => 1,
	    'generate-rasmol-script' => 1,
	    'generate-image'         => 1,
          );


# Get commandline parameters
Getopt::Mixed::init("dir=s
		     project=i
		     image-type=s
		     action=s
                     verbose
		     quiet
		     help
		     d>dir
		     p>project
		     i>image-type
		     a>action
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{'project'}      = $value if($option eq 'project');
	$cfg{'image-type'}   = $value if($option eq 'image-type');
	$cfg{'action'}       = $value if($option eq 'action');
	
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# show what's print()ed without waiting for the newline $|=1;
$|=1;


# check file existance and permissions
if(!-r $cfg{'site-config-file'}) {
	print "Error: Cannot read config.xml file!\n";
	exit 1;
}
if(!-r $cfg{'known-projects-file'} || !-w $cfg{'known-projects-file'}) {
	print "Error: Cannot read and/or write known-projects.xml file!\n";
	exit 1;
}
if(!-r $cfg{'project-images-file'}) {
	print "Error: Cannot read project-images.xml file!\n";
	exit 1;
}


# read the xml files
my $config         = read_xml(%cfg, Name => 'site-config')    || die "Error: Cannot read site-config XML file!\n";
my $known_projects = read_xml(%cfg, Name => 'known-projects') || die "Error: Cannot read known-projects XML file!\n";
my $project_images = read_xml(%cfg, Name => 'project-images') || die "Error: Cannot read project-images XML file!\n";

#print Dumper($config);

# set the path of the xyz2pdb binary
if(exists $config->{settings}->{application_list}->{application}->{xyz2pdb} && 
   exists $config->{settings}->{application_list}->{application}->{xyz2pdb}->{path} &&
   -x $config->{settings}->{application_list}->{application}->{xyz2pdb}->{path}
) {
	$cfg{'xyz2pdb-bin'} = $config->{settings}->{application_list}->{application}->{xyz2pdb}->{path};
}
else {
	print "Error: Cannot find path to the xyz2pdb binary!\n";
	exit 1;
}

# set the path of the rasmol binary 
if(exists $config->{settings}->{application_list}->{application}->{rasmol} && 
   exists $config->{settings}->{application_list}->{application}->{rasmol}->{path} &&
   -x $config->{settings}->{application_list}->{application}->{rasmol}->{path}
) {
	$cfg{'rasmol-bin'} = $config->{settings}->{application_list}->{application}->{rasmol}->{path};
}
else {
	print "Error: Cannot find path to the RasMol binary!\n";
	exit 1;
}

# set the path of the convert binary 
if(exists $config->{settings}->{application_list}->{application}->{convert} && 
   exists $config->{settings}->{application_list}->{application}->{convert}->{path} && 
   -x $config->{settings}->{application_list}->{application}->{convert}->{path}
) {
	$cfg{'convert-bin'} = $config->{settings}->{application_list}->{application}->{convert}->{path};
}
else {
	print "Error: Cannot find path to the convert binary!\n";
	exit 1;
}


# if no action is specified, perform all actions:
#  generate-pdb
#  generate-image (implies generate-rasmol-script)
if($cfg{action} eq '') {
	# if no project is specified, generate for all known-projects which lack images
	if(!$cfg{'project'}) {
		my $projects = 0;
		foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
			update_project_files(%cfg, project => $project, known_projects => $known_projects);
			$projects++;
		}
		if($projects == 0) {
			print "No projects listed yet in the known-projects XML file.\n" if($cfg{verbose});
		}
	}
	else {
		update_project_files(%cfg, known_projects => $known_projects);
	}
}
else {
	if($cfg{'action'} eq 'generate-pdb') {
		# if no project is specified, generate for all known-projects which lack images
		if(!$cfg{'project'}) {
			foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
				generate_pdb(%cfg, project => $project, known_projects => $known_projects);
			}
		}
		else {
			generate_pdb(%cfg, known_projects => $known_projects);
		}
	}
	elsif($cfg{'action'} eq 'generate-rasmol-script') {
		# if no project is specified, generate for all known-projects which lack images
		my ($success, $reason) = (1, '');
		if(!$cfg{'project'}) {
			foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
				generate_rasmol_script_for_project(%cfg, project => $project, known_projects => $known_projects);
			}
		}
		else {
			generate_rasmol_script_for_project(%cfg, known_projects => $known_projects);
		}
	}
	elsif($cfg{'action'} eq 'generate-image') {
		# if no project is specified, generate for all known-projects which lack images
		if(!$cfg{'project'}) {
			foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
				generate_image_for_project(%cfg, project => $project, known_projects => $known_projects);
			}
		}
		else {
			generate_image_for_project(%cfg, known_projects => $known_projects);
		}
	}
	else {
		print "Error: Unknown action: $cfg{action}\n";
	}
}

exit 0; # exit normally

################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disk: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disk: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


##############
#
#     Routine: resize_image()
#  Parameters: source image
#              target image
#              width in pixels
#              height in pixels
#     Returns: 1 on success
#              undef in case of error
# Description: Resize an image with Image::Magick
#

sub resize_image {
        my $source_file  = $_[0];
        my $target_file  = $_[1];
	my $width  	 = $_[2];
	my $height 	 = $_[3];
	my $blur   	 = 1;
	
	if(-r $source_file && (!-e $target_file || -w $target_file)) {
		my $image = Image::Magick->new(magick => 'jpg');

		$image->Read($source_file);

		my ($img_width, $img_height) = $image->Get('width','height');

		if($img_width > $img_height) {
			# Landscape orientated image

			if($img_width > $width) {
				# image is bigger that target size, so resize it
				
				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		elsif($img_height > $img_width) {
			# portrait orientated image
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		else {
			# square image (width eq heigth)
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			elsif($img_width > $width) {
				# image is bigger that target size, so resize it
				
				 my $new_width  = $width;
				 my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			else {
				# image is bigger that target size, so resize it

				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);
				
				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		
		# write resized file to disk
		$image->Write( filename => $target_file );
		chmod(0664, $target_file);
		
		return 1;
	}
	else {
		return undef;
	}
}


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
		
		# allow group users write permissions
		chmod(0664, $args{File});
	}

	return ($success, $reason);
}

sub valid_known_project {
	my $project        = shift;
	my %cfg            = @_;

	print "Validating project: $project ... \n" if($cfg{verbose});

	#my $known_projects = read_xml(%cfg, Name => 'known-projects') 
	#				|| return (0, "Error: Failed to read known-projects XML file!\n");
		
	if($project !~ /^\d+$/) {
		return (0, "Error: Invalid project number! ($project)\n");
	}
	
	if(!exists $known_projects->{known_projects}->{project}->{$project}) {
		return (0, "Error: Project not listed! ($project)\n");
	}

	print "Project validated.\n\n" if($cfg{verbose});

	return 1;
					
}

sub xyz2pdb {
	my %cfg      = @_;
	my $error    = '';

	print "Generating .pdb file from .xyz file.\n" if($cfg{verbose});
	
	if(!$cfg{'xyz-file'} || !-r $cfg{'xyz-file'}) {
		if(!$cfg{'xyz-file'}) {
			$error = "Error: No xyz file specified!\n";
		}
		else {
			$error = "Error: Cannot read xyz file! ($cfg{'xyz-file'})\n";
		}
		
		return(0, $error);
	}
	if(!$cfg{'pdb-file'} || (!-w $cfg{'pdb-file'} && !-w dirname($cfg{'xyz-file'})) ) {
		if(!$cfg{'pdb-file'}) {
			$error = "Error: No pdb file specified!\n";
		}
		elsif(!-w $cfg{'pdb-file'}) {
			$error = "Error: Cannot write pdb file! ($cfg{'pdb-file'})\n";
		}
		elsif(!-w dirname($cfg{'xyz-file'})) {
			$error = "Error: Cannot write project-data directory! (".dirname($cfg{'xyz-file'}).")\n";
		}
		else {
			$error = "Error: Unknown error with pdb-file!\n";
		}

		return(0, $error);
	}
	if(!$cfg{'xyz2pdb-bin'} || !-x $cfg{'xyz2pdb-bin'}) {
		if(!$cfg{'xyz2pdb-bin'}) {
			$error = "Error: No path to xyz2pdb binary specified!\n";
		}
		else {
			$error = "Error: Cannot execute xyz2pdb binary! ($cfg{'xyz2pdb-bin'})\n";
		}
		
		return(0, $error);
	}
	
	if($cfg{verbose}) {
		print "Executing: \"$cfg{'xyz2pdb-bin'}\" -xyzfile \"$cfg{'xyz-file'}\" -o \"$cfg{'pdb-file'}\"\n";
	}
	
	my $return = system( $cfg{'xyz2pdb-bin'},
			     '-xyzfile', $cfg{'xyz-file'},
			     '-o', $cfg{'pdb-file'}
			   );
			     
	if($return == 0) {
		return (1, '');
	}
	else {
		return (0, "Error: xyz2pdb failed!\n");
	}
}

sub generate_pdb {
	my %cfg = @_;

	my ($valid, $error) = valid_known_project($cfg{project}, %cfg);
	if(!$valid) {
		print $error;
		exit 1;
	}
	else {
		# convert .xyz to .pdb if no .pdb exists for this project
		if(!exists $known_projects->{known_projects}->{project}->{$cfg{project}}->{files}->{xyz}) {
			print "Warning: No .xyz file for project $cfg{project}, skipping.\n\n" if($cfg{verbose});
			return;
		}
		
		my $xyz_file = $known_projects->{known_projects}->{project}->{$cfg{project}}->{files}->{xyz};
		my $pdb_file = $xyz_file;
		   $pdb_file =~ s/xyz$/pdb/;
			
		my ($success, $reason) = xyz2pdb( %cfg, 
						  'xyz-file' => $cfg{'www-data-dir'}.$xyz_file,
						  'pdb-file' => $cfg{'www-data-dir'}.$pdb_file
						);
		if(!$success) {
			print $reason;
			exit 1;
		}
		else {
			# save the xyz path in the known-project XML file

			$known_projects = read_xml(%cfg, Name => 'known-projects') 
							|| die "Error: Cannot read known-projects XML file!\n";
	
			my $date = get_date();

			$known_projects->{known_projects}->{project}->{$cfg{project}}->{files}->{pdb}
				= $pdb_file;
			$known_projects->{known_projects}->{project}->{$cfg{project}}->{date_updated}
				= $date;

			my ($success, $reason) = write_xml( %cfg, 
							    Name => 'known-projects',
							    XML  => $known_projects
							  );
			if(!$success) {
				print $reason;
				exit 1;
			}
			else {
				if($cfg{verbose}) {
					print "Successfully converted .xyz to .pdb for project $cfg{project}\n\n";
				}
			}
		}
	}
}


##############
#
#     Routine: generate_rasmol_script()
#  Parameters: %cfg
#     Returns: 1 on success or 0 on error
#              the reason string of the error
#              updated %cfg
# Description: generate an image from a .pdb file using RasMol and ImageMagick
#

sub generate_rasmol_script {
	my %cfg = @_;
	
	print "Generating RasMol script.\n" if($cfg{verbose});

	my ($success, $reason) = (1, '');
	
	# read configurationf file: project-images.xml
	my $xml = read_xml( %cfg,
	                    Name => 'project-images'
	                  ) || die "Failed to read project-images XML file!\n";
	
	# get the default image type if none specified
	if(!$cfg{type} || $cfg{type} =~ /^\s*$/) {
		$cfg{type} = $xml->{project_images}->{default_images}->{location}->{upload}->{type};
	}
	if(!$xml->{project_images}->{image}->{$cfg{type}}) {
		return(0, "Unknown image type: $cfg{type}\n", %cfg);
	}
	if(!$xml->{project_images}->{image}->{$cfg{type}}->{rasmol_template}) {	
		return(0, "No RasMol template listed for type: $cfg{type}!\n", %cfg);
	}
	
	# get the filename of the script template for this type
	my $rasmol_template = $xml->{project_images}->{image}->{$cfg{type}}->{rasmol_template};
	
	print "Using tempate: $rasmol_template\n" if($cfg{verbose});
	
	# create the rasmol script from the template
	if(-r $rasmol_template) {
		my $script = '';
		
		if(! ($script = read_file($rasmol_template, $cfg{verbose})) ) {
			return(0, "Error: While reading RasMol template for type: $cfg{type}\n", %cfg);
		}
		
		print "Generating script commands for type: $cfg{type}\n" if($cfg{verbose});
		if($cfg{type} eq 'still') {
			my $filebase = $cfg{filebase};

			# the rasmol template script
			#
			# spacefill on
			# write ppm image.ppm
			# quit
			
			print "Placing commands in template.\n" if($cfg{verbose});
			if(! ($script =~ s/image/$filebase/) ) {
				return(0, "Error: Failed to replace image filename for type: $cfg{type}\n", %cfg);
			}
		}
		elsif($cfg{type} eq 'rotate') {
			# the molecule should be rotated 360 degrees
			# on the x, y or z axis
			# this should be done in $cfg{image_frames} frames
			# with a delay of $delay ticks between the frames
			# The default ticks-per-second is 100
	
			my $axis   = 'y'; # axis to rotate on
			my $turn   = 6;   # degrees to turn on the axis each time (360 degrees / number-of-frames)
			my $frames = 60;  # number of frames
			my $delay  = 10;  # delay between frames (The default ticks-per-second is 100)
			
			# get the image reference
			my $image = $xml->{project_images}->{image}->{$cfg{type}};
			
			# get the default settings if the value is undefinded or invalid
			my @options = qw/frames delay image_size/;
			foreach(@options) {
				if( !$cfg{$_} || !$image->{setting}->{$_}->{option}->{"$cfg{$_}"} ) {
				  	$cfg{$_} = $image->{setting}->{$_}->{selected_option};
				}
			}
			if(!$cfg{axis} || !$image->{setting}->{rotation_axis}->{option}->{"$cfg{axis}"}) {
				$cfg{axis} = $image->{setting}->{rotation_axis}->{selected_option};
			}

			# calculate the degrees to turn each frame
			if($cfg{frames} && $cfg{frames} =~ /^\d+$/ && $cfg{frames} > 0) {
				$cfg{turn} = 360 / $cfg{frames};
			}

			# use the hardcoded values if the values in the XML file are invalid
			if(!$image->{setting}->{rotation_axis}->{option}->{"$cfg{axis}"}) {
				$cfg{axis}  = $axis;
			}
			if(!$image->{setting}->{delay}->{option}->{"$cfg{delay}"}) {
				$cfg{delay} = $delay;
			}
			if($cfg{turn} <= 0 || $cfg{turn} > 360) {
				$cfg{turn} = $turn;
			}
			
			# multiply delay between frames * 10
			# to convert milliseconds -> ticks
			if($cfg{delay} && $cfg{delay} > 0) {
				$cfg{delay} = 10 * $cfg{delay};
			}
			
			if($cfg{verbose}) {
				print "Rotating image properties:\n";
				print "- Image size                : $cfg{image_size}\n";
				print "- Rotating on axis          : $cfg{axis}\n";
				print "- Degrees to turn each time : $cfg{turn}\n";
				print "- Number of frames          : $cfg{frames}\n";
				print "- Delay between frames      : $cfg{delay}\n";
			}

			my $commands = ''; # commands for the RasMol script
			my $cmdline  = ''; # commandline parameters for Image::Magic convert
			my $round    = 0;  # degrees we have rotated on the axis already
			my $frame    = 0;  # frame number counter
			while($round < 360) {
				my $file = $cfg{filebase}.'-frame'.$frame.'.ppm';
				
				$commands .= "rotate $cfg{axis} $cfg{turn}\n";
				$commands .= "write ppm $file\n";
				
				$cmdline  .= " -delay $cfg{delay}" if($cfg{delay} > 0);
				$cmdline  .= " \"$file\"";
			
				$round += $cfg{turn};
				$frame++;
			}
		
			# place the computed commands in the rasmol script
			# the rasmol template script:
			#
			# set hetero off
			# zoom 100
			# spacefill on
			# COMMANDS
			# quit
			
			print "Placing commands in template.\n" if($cfg{verbose});
			if( !($script =~ s/COMMANDS\n/$commands/) ) {
				return(0, "Error: Failed to replace generated commands in RasMol template for type: $cfg{type}\n", %cfg);
			}
			
			$cfg{cmdline} = $cmdline;
			$cfg{frame}   = $frame;
		}
		elsif($cfg{type} eq 'xyz180') {
			# the molecule should be rotated 180 degrees
			# on the x, y and z axis
			# this should be done in $cfg{image_frames} frames
			# with a delay of $delay ticks between the frames
			# The default ticks-per-second is 100
	
			my $axis   = 'x'; # axis to rotate on
			my $turn   = 9;   # degrees to turn on the axis each time (360 degrees / number-of-frames)
			my $frames = 60;  # number of frames
			my $delay  = 10;  # delay between frames (The default ticks-per-second is 100)
			
			# get the image reference
			my $image = $xml->{project_images}->{image}->{$cfg{type}};
			
			# get the default settings if the value is undefinded or invalid
			my @options = qw/frames delay image_size/;
			foreach(@options) {
				if( !$cfg{$_} || !$image->{setting}->{$_}->{option}->{"$cfg{$_}"} ) {
				  	$cfg{$_} = $image->{setting}->{$_}->{selected_option};
				}
			}

			# calculate the degrees to turn each frame
			if($cfg{frames} && $cfg{frames} =~ /^\d+$/ && $cfg{frames} > 0) {
				$cfg{turn} = (180 * 3) / $cfg{frames};
			}

			# use the hardcoded values if the values in the XML file are invalid
			if(!$image->{setting}->{delay}->{option}->{"$cfg{delay}"}) {
				$cfg{delay} = $delay;
			}
			if($cfg{turn} <= 0 || $cfg{turn} > 180) {
				$cfg{turn} = $turn;
			}
			
			# multiply delay between frames * 10
			# to convert milliseconds -> ticks
			if($cfg{delay} && $cfg{delay} > 0) {
				$cfg{delay} = 10 * $cfg{delay};
			}
			
			if($cfg{verbose}) {
				print "Rotating image properties:\n";
				print "- Image size                : $cfg{image_size}\n";
				print "- Degrees to turn each time : $cfg{turn}\n";
				print "- Number of frames          : $cfg{frames}\n";
				print "- Delay between frames      : $cfg{delay}\n";
			}

			my $commands = ''; # commands for the RasMol script
			my $cmdline  = ''; # commandline parameters for Image::Magic convert
			my $round    = 0;  # degrees we have rotated on the axis already
			my $frame    = 0;  # frame number counter

			$cfg{axis} = 'x';
			while($round < 180) {
				my $file = $cfg{filebase}.'-frame'.$frame.'.ppm';
				
				$commands .= "rotate $cfg{axis} $cfg{turn}\n";
				$commands .= "write ppm $file\n";
				
				$cmdline  .= " -delay $cfg{delay}" if($cfg{delay} > 0);
				$cmdline  .= " \"$file\"";
			
				$round += $cfg{turn};
				$frame++;
			}
			$round = 0;
			$cfg{axis} = 'y';
			while($round < 180) {
				my $file = $cfg{filebase}.'-frame'.$frame.'.ppm';
				
				$commands .= "rotate $cfg{axis} $cfg{turn}\n";
				$commands .= "write ppm $file\n";
				
				$cmdline  .= " -delay $cfg{delay}" if($cfg{delay} > 0);
				$cmdline  .= " \"$file\"";
			
				$round += $cfg{turn};
				$frame++;
			}
			$round = 0;
			$cfg{axis} = 'z';
			while($round < 180) {
				my $file = $cfg{filebase}.'-frame'.$frame.'.ppm';
				
				$commands .= "rotate $cfg{axis} $cfg{turn}\n";
				$commands .= "write ppm $file\n";
				
				$cmdline  .= " -delay $cfg{delay}" if($cfg{delay} > 0);
				$cmdline  .= " \"$file\"";
			
				$round += $cfg{turn};
				$frame++;
			}
		
			# place the computed commands in the rasmol script
			# the rasmol template script:
			#
			# set hetero off
			# zoom 100
			# spacefill on
			# COMMANDS
			# quit
			
			print "Placing commands in template.\n" if($cfg{verbose});
			if( !($script =~ s/COMMANDS\n/$commands/) ) {
				return(0, "Failed to replace generated commands in RasMol template for type: $cfg{type}\n", %cfg);
			}
			
			$cfg{cmdline} = $cmdline;
			$cfg{frame}   = $frame;
		}
		else {
			return(0, "No script handling defined for: $cfg{type}\n", %cfg);
		}
		
		# save the rasmol script
		# and log update

		my $rasmol_script = $cfg{rasmol_script};
		my ($sub_success, $sub_reason) = save_file($rasmol_script, $script, mode => '0664');
			
		if($sub_success) {
			print "Saved RasMol script to: $rasmol_script\n" if($cfg{verbose});
		}
		else {
			return(0, $sub_reason, %cfg);
		}
	}
	else {
		return(0, "Error: Cannot read RasMol template for type: $cfg{type}\n", %cfg);
	}

	return ($success, $reason, %cfg);
}


sub generate_rasmol_script_for_project {
	my %cfg = @_;
	
	my ($valid, $error) = valid_known_project($cfg{project}, %cfg);
	if(!$valid) {
		print $error;
		exit 1;
	}
	
	my ($success, $reason) = (1, '');
	if($cfg{'image-type'} ne '') {
		($success, $reason, %cfg) 
			= generate_rasmol_script( %cfg,
						  type          => $cfg{'image-type'},
						  filebase      => $cfg{'project-data-dir'}.'p'.$cfg{'project'},
						  rasmol_script => $cfg{'project-data-dir'}.'p'.$cfg{project}.'.'.$cfg{'image-type'}.'.script'
						);
		if(!$success) {
			print $reason;
			exit 1;
		}
		else {
			my $file = $cfg{generated_image};
			my $xml = read_xml(%cfg, Name => 'known-projects') 
						|| die "Failed to read known-projects XML file!\n";
			
			$xml->{known_projects}->{project}->{"$cfg{project}"}->{date_updated} 
				= get_date();
			
			$xml->{known_projects}->{project}->{"$cfg{project}"}->{image_list}->{image}->{"$cfg{'image-type'}"}->{script} 
				= $cfg{'project-data-dir'}.'p'.$cfg{project}.'.'.$cfg{'image-type'}.'.script';
			
			# write updated file
			($success, $reason) = write_xml( %cfg,
							 Name => 'known-projects',
							 XML  => $xml,
						       );
									 
			if(!$success) {
				print $reason;
				exit 1;
			}
			else {
				if($cfg{verbose}) {
					print "Successfully generated RasMol script ";
					print "of type $cfg{'image-type'} for project $cfg{project}\n\n";
				}
				$known_projects = $xml;
			}
		}
	}
	else {
		foreach ( sort { $project_images->{project_images}->{image}->{$a}->{order}
				  <=>
				 $project_images->{project_images}->{image}->{$b}->{order}
			       }
			  keys %{$project_images->{project_images}->{image}}
		) {
			$cfg{'image-type'} = $_;
			
			($success, $reason, %cfg) 
				= generate_rasmol_script( %cfg, 
							  type          => $cfg{'image-type'},
							  filebase      => $cfg{'project-data-dir'}.'p'.$cfg{'project'},
							  rasmol_script => $cfg{'project-data-dir'}.'p'.$cfg{project}.'.'.$cfg{'image-type'}.'.script'
							);
			if(!$success) {
				print $reason;
				exit 1;
			}
			else {
				my $file = $cfg{generated_image};
				my $xml = read_xml(%cfg, Name => 'known-projects') 
							|| die "Failed to read known-projects XML file!\n";
				
				$xml->{known_projects}->{project}->{"$cfg{project}"}->{date_updated} 
					= get_date();
				
				$xml->{known_projects}->{project}->{"$cfg{project}"}->{image_list}->{image}->{"$cfg{'image-type'}"}->{script} 
					= $cfg{'project-data-dir'}.'p'.$cfg{project}.'.'.$cfg{'image-type'}.'.script';
				
				# write updated file
				($success, $reason) = write_xml( %cfg,
								 Name => 'known-projects',
								 XML  => $xml,
							       );
										 
				if(!$success) {
					print $reason;
					exit 1;
				}
				else {
					if($cfg{verbose}) {
						print "Successfully generated RasMol script ";
						print "of type $cfg{'image-type'} for project $cfg{project}\n\n";
					}
					$known_projects = $xml;
				}
			}
			print "\n" if($cfg{verbose});
		}
	}	
}	


##############
#
#     Routine: generate_image()
#  Parameters: name of the destination file (without directory or extention)
#              type, optional (if invalid, default will be retrieved from xml file)
#              %cfg
#     Returns: 1 on success
#    	       0 and the reason string on error
# Description: generate an image from a .pdb file using RasMol and ImageMagick
#

sub generate_image {
	my %cfg = @_;

	print "Going to generate an image with RasMol.\n" if($cfg{verbose});

	my ($success, $reason)         = (1, '');
	my ($sub_success, $sub_reason) = (1, '');

	# read configurationf file: project-images.xml
	my $xml = read_xml(%cfg, Name => 'project-images') || die "Failed to read project-images XML file!\n";

	# get the default image type if none specified
	if(!$cfg{type} || $cfg{type} =~ /^\s*$/) {
		$cfg{type} = $xml->{project_images}->{default_images}->{location}->{upload}->{type};
	}
	if(!$xml->{project_images}->{image}->{$cfg{type}}) {
		return(0, "Error: Unknown image type: $cfg{type}\n", %cfg);
	}

	# get the default image size
	if(!$cfg{image_size}) {
		$cfg{image_size} = $xml->{project_images}->{image}->{$cfg{type}}->{setting}->{image_size}->{selected_option};
	}
	if(!$xml->{project_images}->{image}->{$cfg{type}}->{setting}->{image_size}->{option}->{$cfg{image_size}}) {
		return(0, "Error: Invalid image size: $cfg{image_size}", %cfg);
	}

	# check parameters:
	# type           * +
	# filebase         +
	# pdb_file         +
	# rasmol_script    +
	# rasmol-bin       +
	# convert-bin      +
	# frame          * - 
	# cmdline          -
	# image_size     * -
				
	my $pdb_file      = $cfg{pdb_file};
	my $rasmol_script = $cfg{rasmol_script};  # $cfg{filebase}.$cfg{type}.script;
	my $rasmol_bin    = $cfg{'rasmol-bin'};

	# generate the rasmol script for this image type
	print "\n" if($cfg{verbose});
	($sub_success, $sub_reason, %cfg) = generate_rasmol_script(%cfg);
	
	if($sub_success) {
		if($cfg{verbose}) {
			print "\n";
			print "Running RasMol on the .pdb file with the generated script.\n";
		}
		
		if(-r $rasmol_script && -r $pdb_file && -x $cfg{'rasmol-bin'}) {
			# run RasMol with the generated script
			# 
			# Note: 
			# we can't use rasmol's -script option, because that disallows write commands
			# if you give the commands to RasMOl through STDIN you work around this restriction
			
			system("\"$rasmol_bin\" -nodisplay \"$pdb_file\" < \"$rasmol_script\" > /dev/null 2>&1");
			
			if($cfg{type} eq 'still') {
				# resize the image and convert to JPEG in the process
				
				my $ppm_img = $cfg{filebase}.'.ppm';
				my $jpg_img = $cfg{filebase}.'.jpg';
				my ($w, $h) = split /x/, $cfg{image_size};
			
				print "Resizing image.\n" if($cfg{verbose});
				if(resize_image($ppm_img, $jpg_img, $w, $h)) {
					$cfg{generated_image} = $jpg_img;

					print "Removing temporary files.\n" if($cfg{verbose});
					if(-w $ppm_img) {
						unlink $ppm_img || die "Error: Cannot remove file! ($!)\n";
					}
				}
			}
			elsif($cfg{type} eq 'rotate') {
				#print "Checking animation file ... ";
				my $gif_img = $cfg{filebase}.'.gif';
				
				#print "Executing: \"$cfg{'convert-bin'}\" $cfg{cmdline} -geometry $cfg{image_size} $gif_img\n";
				print "Compiling animation with size: $cfg{image_size} to: $gif_img ... " if($cfg{verbose});
				my $convert = `"$cfg{'convert-bin'}" $cfg{cmdline} -geometry $cfg{image_size} $gif_img`;
				print "Done.\n" if($cfg{verbose});
				
				$cfg{generated_image} = $gif_img;
				
				# remove frame files to save space
				print "Removing temporary files ... " if($cfg{verbose});
				while($cfg{frame}--) {
					my $file = $cfg{filebase}.'-frame'.$cfg{frame}.'.ppm';
					if(-w $file) {
						#print "rm $file\n";
						unlink $file || die "Error: Cannot remove file ($!)\n";
					}
				}
				print "Done.\n" if($cfg{verbose});
			}
			elsif($cfg{type} eq 'xyz180') {
				my $gif_img = $cfg{filebase}.'.'.$cfg{type}.'.gif';
				
				#print "Executing: \"$cfg{'convert-bin'}\" $cfg{cmdline} -geometry $size $gif_img\n";
				print "Compiling animation at size: $cfg{image_size} to: $gif_img ... " if($cfg{verbose});
				my $convert = `"$cfg{'convert-bin'}" $cfg{cmdline} -geometry $cfg{image_size} $gif_img`;
				print "Done.\n" if($cfg{verbose});
				
				$cfg{generated_image} = $gif_img;
				
				# remove frame files to save space
				print "Removing temporary files ... " if($cfg{verbose});
				while($cfg{frame}--) {
					my $file = $cfg{filebase}.'-frame'.$cfg{frame}.'.ppm';
					if(-w $file) {
						#print "rm $file\n";
						unlink $file || die "Error: Cannot remove file ($!)\n";
					}
				}
				print "Done.\n" if($cfg{verbose});
			}
			else {
				$success = 0;
				$reason .= "Error: No post-processing for type: $cfg{type}\n";
			}
		}
		else {
			$success = 0;
			if(!-r $pdb_file) {
				$reason .= "Error: Cannot read .pdb file!\n";
			}
			if(!-r $rasmol_script) {
				$reason .= "Error: Cannot read rasmol script! ($rasmol_script)\n";
			}
			if(!-x $rasmol_bin) {
				$reason .= "Error: Cannot execute RasMol binary!\n";
			}
		}
	}
	else {
		$success = 0;
		$reason .= $sub_reason;
	}

	return ($success, $reason, %cfg);
}

sub generate_image_for_project {
	my %cfg = @_;
	
	my ($valid, $error) = valid_known_project($cfg{project}, %cfg);
	if(!$valid) {
		print $error;
		exit 1;
	}
	
	my ($success, $reason) = (1, '');
	if($cfg{'image-type'} ne '') {
		my ($success, $reason) = (1, '');
			
		my $rasmol_script  = $cfg{'project-data-dir'};
		   $rasmol_script .= 'p'.$cfg{project}.'.'.$cfg{'image-type'}.'.script';
	
		my $project_images = $cfg{'project-images-file'};
		if($project_images =~ s/^$cfg{'www-data-dir'}//) {}

		my $pdb_file  = $cfg{'project-data-dir'};
		   $pdb_file .= 'p'.$cfg{project}.'.pdb';
	
		my $rasmol_bin = $cfg{'rasmol-bin'};
        	if($rasmol_bin =~ s/^$cfg{'www-data-dir'}//) {}
		
		my $convert_bin = $cfg{'convert-bin'};
	        if($convert_bin =~ s/^$cfg{'www-data-dir'}//) {}

		($success,
		 $reason,
		 %cfg) = generate_image( %cfg,
		                          type          => $cfg{'image-type'},
					  filebase      => $cfg{'project-data-dir'}.'p'.$cfg{project},
					  rasmol_script => $rasmol_script,
					  pdb_file      => $pdb_file,
					  verbose       => $cfg{verbose},
	
					  'rasmol-bin'          => $rasmol_bin,
					  'convert-bin'         => $convert_bin,
					  'project-images-file' => $project_images
					);
	
		if(!$success) {
			print $reason;
			exit 1;
		}
		else {
			my $file = $cfg{generated_image};
			my $xml = read_xml(%cfg, Name => 'known-projects') 
						|| die "Failed to read known-projects XML file!\n";
			
			$xml->{known_projects}->{project}->{"$cfg{project}"}->{date_updated} 
				= get_date();
			
			$xml->{known_projects}->{project}->{"$cfg{project}"}->{image_list}->{image}->{"$cfg{'image-type'}"}->{script} 
				= $rasmol_script;
			
			$xml->{known_projects}->{project}->{"$cfg{project}"}->{image_list}->{image}->{"$cfg{'image-type'}"}->{file} 
				= $file;
			
			$xml->{known_projects}->{project}->{"$cfg{project}"}->{files}->{pdb}
				= $pdb_file;
					
			# write updated file
			($success, $reason) = write_xml( %cfg,
							 Name => 'known-projects',
							 XML  => $xml,
						       );
									 
			if(!$success) {
				print $reason;
				exit 1;
			}
			else {
				if($cfg{verbose}) {
					print "Successfully generated image of type $cfg{'image-type'} for project $cfg{project}\n\n";
				}
				$known_projects = $xml;
			}
		}
	}
	else {
		foreach ( sort { $project_images->{project_images}->{image}->{$a}->{order}
				  <=>
				 $project_images->{project_images}->{image}->{$b}->{order}
			       }
			  keys %{$project_images->{project_images}->{image}}
		) {
			$cfg{'image-type'} = $_;
			
			my ($success, $reason) = (1, '');
			
			my $rasmol_script  = $cfg{'project-data-dir'};
			   $rasmol_script .= 'p'.$cfg{project}.'.'.$cfg{'image-type'}.'.script';
	
			my $project_images = $cfg{'project-images-file'};
			if($project_images =~ s/^$cfg{'www-data-dir'}//) {}
	
			my $pdb_file  = $cfg{'project-data-dir'};
			   $pdb_file .= 'p'.$cfg{project}.'.pdb';
		
			my $rasmol_bin = $cfg{'rasmol-bin'};
	        	if($rasmol_bin =~ s/^$cfg{'www-data-dir'}//) {}
			
			my $convert_bin = $cfg{'convert-bin'};
		        if($convert_bin =~ s/^$cfg{'www-data-dir'}//) {}
	
			($success,
			 $reason,
			 %cfg) = generate_image( %cfg,
			                          type          => $cfg{'image-type'},
						  filebase      => $cfg{'project-data-dir'}.'p'.$cfg{project},
						  rasmol_script => $rasmol_script,
						  pdb_file      => $pdb_file,
						  verbose       => $cfg{verbose},
		
						  'rasmol-bin'          => $rasmol_bin,
						  'convert-bin'         => $convert_bin,
						  'project-images-file' => $project_images
						);
		
			if(!$success) {
				print $reason;
				exit 1;
			}
			else {
				my $file = $cfg{generated_image};
				my $xml = read_xml(%cfg, Name => 'known-projects') 
							|| die "Failed to read known-projects XML file!\n";
				
				$xml->{known_projects}->{project}->{"$cfg{project}"}->{date_updated} 
					= get_date();
				
				$xml->{known_projects}->{project}->{"$cfg{project}"}->{image_list}->{image}->{"$cfg{'image-type'}"}->{script} 
					= $rasmol_script;
				
				$xml->{known_projects}->{project}->{"$cfg{project}"}->{image_list}->{image}->{"$cfg{'image-type'}"}->{file} 
					= $file;
				
				$xml->{known_projects}->{project}->{"$cfg{project}"}->{files}->{pdb}
					= $pdb_file;
						
				# write updated file
				($success, $reason) = write_xml( %cfg,
								 Name => 'known-projects',
								 XML  => $xml,
							       );
										 
				if(!$success) {
					print $reason;
					exit 1;
				}
				else {
					if($cfg{verbose}) {
						print "Successfully generated image of type $cfg{'image-type'} for project $cfg{project}\n\n";
					}
					$known_projects = $xml;
				}
			}
			print "\n" if($cfg{verbose});
		}
	}
}

sub update_project_files {
	my %cfg = @_;

	my ($valid, $error) = valid_known_project($cfg{project}, %cfg);
	if(!$valid) {
		print $error;
		exit 1;
	}
	else {
		# convert .xyz to .pdb if no .pdb exists for this project
		if(!exists $known_projects->{known_projects}->{project}->{$cfg{project}}->{files}->{xyz}) {
			print "Warning: No .xyz file for project: $cfg{project}, skipping.\n\n" if($cfg{verbose});
			return;
		}
		
		my $xyz_file = $known_projects->{known_projects}->{project}->{$cfg{project}}->{files}->{xyz};
		my $pdb_file = $xyz_file;
		   $pdb_file =~ s/xyz$/pdb/;
			
		if(!exists $known_projects->{known_projects}->{project}->{$cfg{project}}->{files}->{pdb}) {
			generate_pdb(%cfg);
		}

		# generate the RasMol scripts and images for all missing images types
		foreach(keys %{$project_images->{project_images}->{image}}) {
			$cfg{'image-type'} = $_;
			
			if(!exists $known_projects->{known_projects}->{project}->{$cfg{project}}->{image_list}->{image}->{"$cfg{'image-type'}"}->{file}) {
				my ($success, $reason) = generate_image_for_project(%cfg);
				if(!$success) {
					print $reason;
					exit 1;
				}
			}
		}
	}
}


						
sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>         Path of the FCI server directory\n";
	print "-p, --project <NUMBER>  Project number to update\n";
	print "-i, --image-type <TYPE> Image type to generate\n";
	print "-a, --action <ACTION>   Action to perform on the project:\n";
	print "                         generate-pdb           (Convert .xyz to .pdb)\n";
	print "                         generate-rasmol-script (Create RasMol script)\n";
	print "                         generate-image         (Create image(s) and RasMol script(s))\n";
	print "-v, --verbose           Enable verbose output\n";
	print "-q, --quiet             Don't print any output at all, except for errors\n";
	print "-h, --help              Show this usage information\n";
}
